@@ -1,22 +1,23 @@
 /*
  *      fm-dummy-monitor.c
  *
- *      Copyright 2009 PCMan <pcman@debian>
+ *      Copyright 2009 PCMan <pcman.tw@gmail.com>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of the Libfm library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -24,7 +25,7 @@
  * @short_description: Replacement of #GFileMonitor for virtual filesystems.
  * @title: FmDummyMonitor
  *
- * @include: libfm/fm-dummy-monitor.h
+ * @include: libfm/fm.h
  *
  * The #FmDummyMonitor represents dummy #GFileMonitor variant which does
  * not monitor any files but can be used as monitor object.
@@ -36,7 +37,7 @@
 
 G_DEFINE_TYPE(FmDummyMonitor, fm_dummy_monitor, G_TYPE_FILE_MONITOR);
 
-static gboolean cancel()
+static gboolean cancel(GFileMonitor *monitor)
 {
     return TRUE;
 }
