@@ -1,22 +1,23 @@
 /*
  *      fm-mime-type.h
- *      
+ *
  *      Copyright 2009 - 2012 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
- *      
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
- *      
- *      This program is distributed in the hope that it will be useful,
+ *
+ *      This file is a part of the Libfm library.
+ *
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
- *      
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
+ *
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifndef _FM_MIME_TYPE_H_
@@ -48,7 +49,8 @@ FmMimeType* fm_mime_type_from_name(const char* type);
 
 FmMimeType* _fm_mime_type_get_inode_directory();
 FmMimeType* _fm_mime_type_get_inode_x_shortcut();
-FmMimeType* _fm_mime_type_get_inode_x_mountable();
+FmMimeType* _fm_mime_type_get_inode_mount_point();
+FmMimeType* _fm_mime_type_get_application_x_desktop();
 
 FmMimeType* fm_mime_type_ref(FmMimeType* mime_type);
 void fm_mime_type_unref(gpointer mime_type_);
@@ -63,7 +65,10 @@ const char* fm_mime_type_get_desc(FmMimeType* mime_type);
 
 /* Get installed external thumbnailers for the mime-type.
  * Returns a list of FmThumbnailer. */
+#ifndef FM_DISABLE_DEPRECATED
 const GList* fm_mime_type_get_thumbnailers(FmMimeType* mime_type);
+#endif
+GList* fm_mime_type_get_thumbnailers_list(FmMimeType* mime_type);
 
 void fm_mime_type_add_thumbnailer(FmMimeType* mime_type, gpointer thumbnailer);
 
