@@ -1,22 +1,24 @@
 /*
  *      fm-nav-history.c
- *      
+ *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
- *      
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
- *      
- *      This program is distributed in the hope that it will be useful,
+ *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *
+ *      This file is a part of the Libfm library.
+ *
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
- *      
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
+ *
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -24,7 +26,7 @@
  * @short_description: Simple navigation history management.
  * @title: FmNavHistory
  *
- * @include: libfm/fm-nav-history.h
+ * @include: libfm/fm.h
  *
  * The #FmNavHistory object implements history for paths that were
  * entered in some input bar and allows to add, remove or move items in it.
@@ -32,6 +34,20 @@
 
 #include "fm-nav-history.h"
 
+struct _FmNavHistory
+{
+    GObject parent;
+    GQueue items;
+    GList* cur;
+    gint n_max;
+    guint n_cur;
+};
+
+struct _FmNavHistoryClass
+{
+    GObjectClass parent_class;
+};
+
 static void fm_nav_history_finalize (GObject *object);
 
 G_DEFINE_TYPE(FmNavHistory, fm_nav_history, G_TYPE_OBJECT);
@@ -96,7 +112,7 @@ FmNavHistory *fm_nav_history_new(void)
  *
  * Since: 0.1.0
  *
- * Deprecated: 1.0.2: Don't use it in newer code.
+ * Deprecated: 1.0.2:
  */
 const GList* fm_nav_history_list(FmNavHistory* nh)
 {
@@ -133,7 +149,7 @@ const FmNavHistoryItem* fm_nav_history_get_cur(FmNavHistory* nh)
  *
  * Since: 0.1.0
  *
- * Deprecated: 1.0.2: Don't use it in newer code.
+ * Deprecated: 1.0.2:
  */
 const GList* fm_nav_history_get_cur_link(FmNavHistory* nh)
 {
