@@ -1,22 +1,25 @@
 /*
  *      fm-path.c
  *
- *      Copyright 2009 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2009-2014 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2009 Juergen Hoetzel <juergen@archlinux.org>
+ *      Copyright 2012-2013 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of the Libfm library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -24,7 +27,7 @@
  * @short_description: Path representation for libfm.
  * @title: FmPath
  *
- * @include: libfm/fm-path.h
+ * @include: libfm/fm.h
  *
  */
 
@@ -41,10 +44,15 @@
 #include <limits.h>
 #include <glib/gi18n-lib.h>
 
+#define BASENAME_AS_DISP_NAME ((char *)-1)
+
 struct _FmPath
 {
     gint n_ref;
     FmPath* parent;
+    char *disp_name;
+    GSequenceIter *iter; /* iterator in parent, NULL for root path */
+    GSequence *children; /* children to reuse paths */
     guchar flags; /* FmPathFlags flags : 8; */
     char name[1]; /* basename: in local encoding if native, uri-escaped otherwise */
 };
@@ -70,6 +78,12 @@ static FmPath* network_root = NULL;*/
 
 static FmPath* apps_root_path = NULL;
 
+static GSList* roots = NULL;
+
+/* a lock for access to changeable data: roots list, and
+   members of FmPath struct: disp_name, iter, children */
+G_LOCK_DEFINE_STATIC(roots);
+
 static FmPath* _fm_path_alloc(FmPath* parent, int name_len, int flags)
 {
     FmPath* path;
@@ -77,6 +91,9 @@ static FmPath* _fm_path_alloc(FmPath* parent, int name_len, int flags)
     path->n_ref = 1;
     path->flags = flags;
     path->parent = parent ? fm_path_ref(parent) : NULL;
+    path->disp_name = NULL;
+    path->children = NULL;
+    path->iter = NULL;
     return path;
 }
 
@@ -85,6 +102,15 @@ static inline FmPath* _fm_path_new_internal(FmPath* parent, const char* name, in
     FmPath* path = _fm_path_alloc(parent, name_len, flags);
     memcpy(path->name, name, name_len);
     path->name[name_len] = '\0';
+    if (parent)
+    {
+        G_LOCK(roots);
+        if (parent->children == NULL)
+            parent->children = g_sequence_new(NULL);
+        path->iter = g_sequence_insert_sorted(parent->children, path,
+                                              (GCompareDataFunc)fm_path_compare, NULL);
+        G_UNLOCK(roots);
+    }
     return path;
 }
 
@@ -100,10 +126,12 @@ static inline FmPath* _fm_path_new_internal(FmPath* parent, const char* name, in
 static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** remaining)
 {
     FmPath* path;
+    GSList* l;
     char* buf;
     const char* uri_end = uri + len;
     const char* host;
     const char* host_end;
+    const char *disp_name = NULL;
     int scheme_len, host_len;
     int flags;
 
@@ -125,12 +153,13 @@ static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** rema
 
     /* now there should be // following :, or no slashes at all, such as mailto:someone@somewhere.net */
     host = uri + scheme_len + 1;
-    while(*host == '/' && host < uri_end) /* skip the slashes */
-        ++host;
+    if(host + 2 < uri_end && host[0] == '/' && host[1] == '/')
+        host += 2;
 
     flags = 0;
     host_len = 0;
-    if(scheme_len == 4 && g_ascii_strncasecmp(uri, "file", 4) == 0) /* handles file:/// */
+    if(scheme_len == 4 && g_ascii_strncasecmp(uri, "file", 4) == 0 &&
+       host[0] == '/') /* handle file:/ or file:///, file://host/ isn't our */
     {
         if(remaining)
             *remaining = host;
@@ -147,11 +176,13 @@ static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** rema
     {
         flags |= FM_PATH_IS_VIRTUAL; /* FIXME: deprecated */
         host_end = host;
+        disp_name = _("Devices");
     }
     else if(scheme_len == 7 && g_ascii_strncasecmp(uri, "network", 7) == 0)
     {
         flags |= FM_PATH_IS_VIRTUAL; /* FIXME: deprecated */
         host_end = host;
+        disp_name = _("Network");
     }
     else if(scheme_len == 6 && g_ascii_strncasecmp(uri, "mailto", 6) == 0)
     {
@@ -167,13 +198,10 @@ static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** rema
         while(host_end < uri_end && *host_end != '/') /* find the end of host name */
             ++host_end;
         host_len = (host_end - host);
-        /* FIXME: is this ever needed? VFS should handle it */
         if(scheme_len == 4 && g_ascii_strncasecmp(uri, "menu", 4) == 0)
         {
             if(host_len == 0) /* fallback to applications */
             {
-                host = "applications";
-                host_len = 12;
                 if(remaining)
                     *remaining = uri_end;
                 return fm_path_ref(apps_root_path);
@@ -193,7 +221,22 @@ static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** rema
 
     /* it's reasonable to have double slashes :// for URIs other than mailto: */
     len = scheme_len + 3 + host_len + 1;
+    G_LOCK(roots);
+    for(l = roots; l; l = l->next)
+    {
+        path = l->data;
+        if(strncmp(path->name, uri, scheme_len) == 0 &&
+           (!host_len || !strncmp(&path->name[scheme_len + 3], host, host_len)) &&
+           strcmp(&path->name[len-1], "/") == 0)
+        {
+            fm_path_ref(path);
+            G_UNLOCK(roots);
+            return path;
+        }
+    }
     path = _fm_path_alloc(NULL, len, flags);
+    roots = g_slist_append(roots, path);
+    G_UNLOCK(roots);
     buf = path->name;
     memcpy(buf, uri, scheme_len); /* the scheme */
     buf += scheme_len;
@@ -206,36 +249,16 @@ static FmPath* _fm_path_new_uri_root(const char* uri, int len, const char** rema
     }
     buf[0] = '/'; /* the trailing / */
     buf[1] = '\0';
-
+    if (disp_name)
+        path->disp_name = g_strdup(disp_name); /* no lock required, it's new data */
     return path;
 
 on_error: /* this is not a valid URI */
-    /* FIXME: should we return root or NULL? */
     if(remaining)
         *remaining = uri + len;
     return fm_path_ref(root_path);
 }
 
-#if 0
-static inline FmPath* _fm_path_reuse_existing_paths(FmPath* parent, const char* basename, int name_len)
-{
-    FmPath* current;
-    /* This is a way to reuse cached FmPath objects created for $HOME and desktop dir.
-     * Since most of the files a user may use are under $HOME, reusing this can
-     * more or less reduce memory usage. However, this may slow things down a little. */
-    for(current = desktop_path; current; current = current->parent)
-    {
-        if(fm_path_equal(current->parent, parent))
-        {
-            if(strncmp(basename, current->name, name_len) == 0 && current->name[name_len] == '\0')
-                return fm_path_ref(current);
-            break;
-        }
-    }
-    return NULL;
-}
-#endif
-
 /**
  * fm_path_new_child_len
  * @parent: (allow-none): a parent path
@@ -251,11 +274,11 @@ static inline FmPath* _fm_path_reuse_existing_paths(FmPath* parent, const char*
  * Returns: (transfer full): a new #FmPath for the path. You have to call
  * fm_path_unref() when it's no longer needed.
  */
+static
 FmPath* _fm_path_new_child_len(FmPath* parent, const char* basename, int name_len,
-                               gboolean dont_escape)
+                               gboolean dont_escape, gboolean is_query)
 {
     FmPath* path;
-    gboolean append_slash = FALSE;
     int flags;
 
     /* skip empty basename */
@@ -264,37 +287,22 @@ FmPath* _fm_path_new_child_len(FmPath* parent, const char* basename, int name_le
 
     if(G_LIKELY(parent)) /* remove slashes if needed. */
     {
-        #if 0 /* This saves some memory, but it's too inefficient. */
-        /* Optimization: reuse existing FmPaths */
-        if(fm_path_is_native(parent))
-        {
-            path = _fm_path_reuse_existing_paths(parent, basename, name_len);
-            if(G_UNLIKELY(path))
-                return path;
-        }
-        #endif
-
         flags = parent->flags; /* inherit flags of parent */
-        while(basename[0] == '/')
+        if (!is_query) while(basename[0] == '/')
         {
             ++basename;
             --name_len;
         }
-        while(name_len > 0 && basename[name_len-1] == '/')
+        if (!is_query) while(name_len > 0 && basename[name_len-1] == '/')
             --name_len;
 
         /* special case for . and .. */
-        if(basename[0] == '.' && (name_len == 1 || (name_len == 2 && basename[1] == '.')))
+        if(basename[0] == '.')
         {
             if(name_len == 1) /* . */
-                return parent ? fm_path_ref(parent) : NULL;
-            else /* .. */
-            {
-                if(parent)
-                    return parent->parent ? fm_path_ref(parent->parent) : fm_path_ref(parent);
-                else
-                    return NULL;
-            }
+                return fm_path_ref(parent);
+            else if(name_len == 2 && basename[1] == '.') /* .. */
+                return parent->parent ? fm_path_ref(parent->parent) : fm_path_ref(parent);
         }
     }
     else /* this basename is root of the fs (no parent), it can be "/" or something like "ftp://user@host/" */
@@ -307,43 +315,80 @@ FmPath* _fm_path_new_child_len(FmPath* parent, const char* basename, int name_le
     }
 
     /* remove tailing slashes */
-    while(name_len > 0 && basename[name_len-1] == '/')
+    if (!is_query) while(name_len > 0 && basename[name_len-1] == '/')
         --name_len;
 
     if(name_len == 0)
-        return parent ? fm_path_ref(parent) : NULL;
+        return fm_path_ref(parent);
 
+    /* create path data first */
     if(dont_escape)
     {
-        path = _fm_path_alloc(parent, (G_UNLIKELY(append_slash) ? name_len + 1 : name_len), flags);
+        path = _fm_path_alloc(parent, name_len, flags);
         memcpy(path->name, basename, name_len);
     }
     else
     {
         GString *str = g_string_new_len(basename, name_len);
         /* remote file names don't come escaped from gvfs; isn't that a bug of gvfs? */
-        char *escaped = g_uri_escape_string(str->str, "/", TRUE);
+        char *escaped = g_uri_escape_string(str->str, G_URI_RESERVED_CHARS_ALLOWED_IN_PATH, FALSE);
         /* g_debug("got child %s", escaped); */
         name_len = strlen(escaped);
-        path = _fm_path_alloc(parent, (G_UNLIKELY(append_slash) ? name_len + 1 : name_len), flags);
+        path = _fm_path_alloc(parent, name_len, flags);
         memcpy(path->name, escaped, name_len);
         g_free(escaped);
         g_string_free(str, TRUE);
     }
-    if(G_UNLIKELY(append_slash))
+    path->name[name_len] = '\0';
+
+    G_LOCK(roots);
+    if (parent->children)
     {
-        path->name[name_len] = '/';
-        path->name[name_len + 1] = '\0';
+        GSequenceIter *iter;
+
+        /* try to reuse existing path */
+#if GLIB_CHECK_VERSION(2, 28, 0)
+        iter = g_sequence_lookup(parent->children, path,
+                                 (GCompareDataFunc)fm_path_compare, NULL);
+#else
+        iter = g_sequence_search(parent->children, path,
+                                 (GCompareDataFunc)fm_path_compare, NULL);
+        if (iter)
+        {
+            FmPath *pp = NULL;
+            iter = g_sequence_iter_prev(iter);
+            if (iter)
+                pp = g_sequence_get(iter);
+            if (fm_path_compare(pp, path) != 0)
+                iter = NULL;
+        }
+#endif
+        if (iter)
+        {
+                /* g_debug("found reusable path '%.*s'", name_len, basename); */
+            FmPath *np = fm_path_ref(g_sequence_get(iter));
+            G_UNLOCK(roots); /* we should not unref with lock up */
+            fm_path_unref(path); /* drop this path and reuse found one */
+            return np;
+        }
+        else
+            path->iter = g_sequence_insert_sorted(parent->children, path,
+                                                  (GCompareDataFunc)fm_path_compare,
+                                                  NULL);
     }
     else
-        path->name[name_len] = '\0';
+    {
+        parent->children = g_sequence_new(NULL);
+        path->iter = g_sequence_append(parent->children, path);
+    }
+    G_UNLOCK(roots);
     return path;
 }
 
 FmPath* fm_path_new_child_len(FmPath* parent, const char* basename, int name_len)
 {
     return _fm_path_new_child_len(parent, basename, name_len,
-                                  parent && fm_path_is_native(parent));
+                                  parent && fm_path_is_native(parent), FALSE);
 }
 
 /**
@@ -364,7 +409,7 @@ FmPath* fm_path_new_child(FmPath* parent, const char* basename)
     {
         int baselen = strlen(basename);
         return _fm_path_new_child_len(parent, basename, baselen,
-                                      parent && fm_path_is_native(parent));
+                                      parent && fm_path_is_native(parent), FALSE);
     }
     return G_LIKELY(parent) ? fm_path_ref(parent) : NULL;
 }
@@ -419,6 +464,7 @@ FmPath* fm_path_new_relative(FmPath* parent, const char* rel)
     if(G_UNLIKELY(!rel || !*rel)) /* relative path is empty */
         return parent ? fm_path_ref(parent) : fm_path_ref(root_path); /* return parent */
 
+    /* g_debug("fm_path_new_relative: %s", rel); */
     if(G_LIKELY(parent))
     {
         char* sep;
@@ -429,27 +475,16 @@ FmPath* fm_path_new_relative(FmPath* parent, const char* rel)
             path = fm_path_ref(parent);
         else
         {
-#if 0       /* FIXME: Let's optimize this later. Make things working first is more important. */
-            /* use some pre-defined paths when possible */
-            if(G_UNLIKELY(parent == root_path))
-            {
-                if(strcmp(home_dir + 1, rel) == 0)
-                    return fm_path_ref(home_path);
-                if(strcmp(desktop_dir + 1, rel) == 0)
-                    return fm_path_ref(desktop_dir);
-            }
-#endif
-
             sep = strchr(rel, '/');
             if(sep)
             {
-                FmPath* new_parent = _fm_path_new_child_len(parent, rel, sep - rel, TRUE);
+                FmPath *new_parent = _fm_path_new_child_len(parent, rel, sep - rel, TRUE, FALSE);
                 path = fm_path_new_relative(new_parent, sep + 1);
                 fm_path_unref(new_parent);
             }
             else
             {
-                path = _fm_path_new_child_len(parent, rel, strlen(rel), TRUE);
+                path = _fm_path_new_child_len(parent, rel, strlen(rel), TRUE, FALSE);
             }
         }
     }
@@ -494,6 +529,9 @@ FmPath* fm_path_new_for_path(const char* path_name)
  * You can call fm_path_to_uri() to convert a FmPath to a escaped URI
  * string.
  *
+ * This function never fail but it may return root path if URI is local
+ * and contains illegal characters in path elements such as '/'.
+ *
  * Returns: (transfer full): a new #FmPath for the @uri.
  */
 FmPath* fm_path_new_for_uri(const char* uri)
@@ -508,12 +546,15 @@ FmPath* fm_path_new_for_uri(const char* uri)
     {
         if(root == root_path)
         {
-            /* handle file:// URIs */
+            /* handle file:/// URIs */
             char *filename = g_filename_from_uri(uri, NULL, NULL);
-            /* FIXME: handle inconvertable URIs - its translated into '/' now */
+            /* NOTE: g_filename_from_uri() will fail on unescaping '%2F' */
             path = fm_path_new_relative(root, filename);
             g_free(filename);
         }
+        else if (strchr(rel_path, '?') != NULL) /* it's query not regular URI */
+            path = _fm_path_new_child_len(root, rel_path+1, strlen(rel_path)-1,
+                                          TRUE, TRUE);
         else
             path = fm_path_new_relative(root, rel_path);
         fm_path_unref(root);
@@ -523,46 +564,157 @@ FmPath* fm_path_new_for_uri(const char* uri)
     return path;
 }
 
-#ifndef FM_DISABLE_DEPRECATED
+/* looks for known display_name, returns referenced child */
+static FmPath *_lookup_in_children(FmPath *path, const char *display_name)
+{
+    FmPath *subpath = NULL;
+
+    G_LOCK(roots);
+    if (path->children != NULL)
+    {
+        GSequenceIter *iter = g_sequence_get_begin_iter(path->children);
+        const char *name;
+
+        while (!g_sequence_iter_is_end(iter))
+        {
+            path = (FmPath*)g_sequence_get(iter);
+            name = path->disp_name;
+            if (name)
+            {
+                if (name == BASENAME_AS_DISP_NAME)
+                    name = path->name;
+                if (strcmp(display_name, name) == 0)
+                {
+                    subpath = fm_path_ref(path);
+                    break;
+                }
+            }
+            iter = g_sequence_iter_next(iter);
+        }
+    }
+    G_UNLOCK(roots);
+    return subpath;
+}
+
 /**
  * fm_path_new_for_display_name
- * @path_name: a UTF-8 encoded display name for the path
- * It can either be a POSIX path in UTF-8 encoding, or an unescaped URI
- * (can contain non-ASCII characters and spaces)
+ * @path_name: an UTF-8 encoded display path name
  *
- * You can call fm_path_display_name() to convert a FmPath to a
- * UTF-8 encoded name ready for being displayed in the GUI.
+ * The @path_name can either be a POSIX path in UTF-8 encoding, or an
+ * unescaped URI (can contain non-ASCII characters and spaces). Returned
+ * data should be freed with fm_path_unref() when it's no longer needed.
+ * Display name may be fully unrelated to it's path name and built path
+ * may be inexistant. Use FmJob to check and validate its existance.
  *
- * Returns: a newly created FmPath for the path. You have to call
- * fm_path_unref() when it's no longer needed.
+ * Returns: a newly created #FmPath.
+ *
+ * Since: 0.1.14
  */
-/* FIXME: this is completely invalid way to do this. Display name may be
-   fully unrelated to it's path name.
-   The only correct way is to use g_file_get_child_for_display_name() */
 FmPath* fm_path_new_for_display_name(const char* path_name)
 {
-    FmPath* path;
+    FmPath *path;
+    FmPath *cpath;
+    GFile *file = NULL, *child;
+    char *path_copy, *c, *sep;
+    char sep_char;
+
     if(!path_name || !*path_name || (path_name[0]=='/' && path_name[1] == '\0') )
         return fm_path_ref(root_path);
-    if(path_name[0] == '/') /* native path */
+
+    if (path_name[0] != '/') /* it's an URI */
     {
-        char* filename = g_filename_from_utf8(path_name, -1, NULL, NULL, NULL);
-        if(filename) /* convert from utf-8 to local encoding */
+        sep = strchr(path_name, ':');
+        if (!sep) /* invalid path */
+            /* FIXME: fail on this */
+            return fm_path_ref(root_path);
+        /* FIXME: should we test for "C:\path..." too? */
+        if (sep[1] != '/') /* something like mailto:xx@yy */
+            return fm_path_new_for_uri(path_name);
+        if (sep[2] != '/') /* malformed URI */
+            return fm_path_new_for_str(path_name); /* FIXME: handle this? */
+        path_copy = g_strdup(path_name);
+        c = path_copy + (sep + 3 - path_name); /* set after xx:// */
+        sep = strchr(c, '/');
+        if (sep)
         {
-            path = fm_path_new_for_path(filename);
-            g_free(filename);
+            if (strchr(sep, '?') != NULL)
+            {
+                /* it is not URI path but rather URI query, pass it "as is" */
+                path = fm_path_new_for_uri(path_copy);
+                goto _finish;
+            }
+            /* terminate root URI after third slash */
+            sep++;
+        }
+        /* use c for escaped copy of root path temporarily
+           we use another set of allowed chars, see bug #3599638 */
+        c = g_uri_escape_string(path_copy,
+                                G_URI_RESERVED_CHARS_GENERIC_DELIMITERS
+                                G_URI_RESERVED_CHARS_SUBCOMPONENT_DELIMITERS,
+                                FALSE);
+        path = _fm_path_new_uri_root(c, strlen(c), NULL);
+        /* g_debug("** display URI root: %s",path->name); */
+        g_free(c);
+        c = sep;
+        sep_char = '/'; /* separator for URI is always slash */
+    }
+    else
+    {
+        path_copy = g_strdup(path_name);
+        path = fm_path_ref(root_path);
+        c = path_copy + 1;
+        sep_char = G_DIR_SEPARATOR;
+    }
+    for ( ; c; c = sep)
+    {
+        sep = strchr(c, sep_char);
+        if (sep)
+            *sep++ = '\0'; /* separate one part of path */
+        /* g_debug("** next subpath: %s",c); */
+        if (*c == '\0') /* duplicate '/' */
+            continue;
+        if (strcmp(c, ".") == 0) /* skip "." */
+            continue;
+        /* check if we have it already cached */
+        cpath = _lookup_in_children(path, c);
+        if (cpath)
+        {
+            /* g_debug("** found child: %s => %s",c,cpath->name); */
+        }
+        else if (strcmp(c, "..") == 0) /* go back one dir */
+        {
+            if (path->parent == NULL) /* attempt to go below root */
+                break;
+            cpath = fm_path_ref(path->parent);
         }
         else
-            path = fm_path_ref(root_path);
-    }
-    else /* this is an URI */
-    {
-        /* UTF-8 should be allowed, I think. */
-        path = fm_path_new_for_uri(path_name);
+        {
+            file = fm_path_to_gfile(path);
+            child = g_file_get_child_for_display_name(file, c, NULL);
+            g_object_unref(file);
+            if (child) /* successfull conversion, use name we got */
+            {
+                /* char *cc = g_file_get_uri(child);
+                g_debug("** resolved display name: %s => %s",c,cc);
+                g_free(cc); */
+                c = g_file_get_basename(child);
+            }
+            /* otherwise use basename as fallback */
+            cpath = fm_path_new_child(path, c);
+            /* g_debug("** adding child: %s",cpath->name); */
+            if (child)
+            {
+                g_free(c);
+                g_object_unref(child);
+            }
+        }
+        fm_path_unref(path);
+        path = cpath;
     }
+_finish:
+    g_free(path_copy);
     return path;
 }
-#endif /* FM_DISABLE_DEPRECATED */
 
 /**
  * fm_path_new_for_str
@@ -586,7 +738,6 @@ FmPath* fm_path_new_for_str(const char* path_str)
         return fm_path_ref(root_path);
     if(path_str[0] == '/')
         return fm_path_new_for_path(path_str);
-    /* FIXME: add a support for relative path from FmPathEntry */
     /* UTF-8 should be allowed, I think. */
     escaped = g_uri_escape_string(path_str,
                                   G_URI_RESERVED_CHARS_GENERIC_DELIMITERS
@@ -644,8 +795,28 @@ void fm_path_unref(FmPath* path)
     /* g_debug("fm_path_unref: %s, n_ref = %d", fm_path_to_str(path), path->n_ref); */
     if(g_atomic_int_dec_and_test(&path->n_ref))
     {
+        G_LOCK(roots);
         if(G_LIKELY(path->parent))
+        {
+            if (G_LIKELY(path->iter))
+                g_sequence_remove(path->iter);
+                /* otherwise it's fresh abandoned one, see above */
+            G_UNLOCK(roots); /* we should not unref with lock up */
             fm_path_unref(path->parent);
+        }
+        else
+        {
+            roots = g_slist_remove(roots, path);
+            G_UNLOCK(roots);
+        }
+        /* no further lock required, it is removed from lists */
+        if (path->disp_name != BASENAME_AS_DISP_NAME)
+            g_free(path->disp_name);
+        if (G_UNLIKELY(path->children))
+        {
+            g_assert(g_sequence_get_length(path->children) == 0);
+            g_sequence_free(path->children);
+        }
         g_free(path);
     }
 }
@@ -666,6 +837,26 @@ FmPath* fm_path_get_parent(FmPath* path)
     return path->parent;
 }
 
+/**
+ * fm_path_get_scheme_path
+ * @path: a path
+ *
+ * Retrieves a root path for scheme used by @path. If @path is a local
+ * path then returned data are the same as returned with fm_path_get_root().
+ * Returned data are owned by @path and should be not freed by caller.
+ *
+ * Returns: (transfer none): path of root of @path scheme.
+ *
+ * Since: 1.2.0
+ */
+FmPath* fm_path_get_scheme_path(FmPath* path)
+{
+    g_return_val_if_fail(path != NULL, NULL);
+    while(path->parent != NULL)
+        path = path->parent;
+    return path;
+}
+
 /**
  * fm_path_get_basename
  * @path: a path
@@ -746,12 +937,12 @@ static gchar* fm_path_to_str_int(FmPath* path, gchar** ret, gint str_len)
  *
  * Creates string representation of @path. It can be either file path in
  * local encoding or URI with non-ASCII characters escaped (RFC 3986).
+ * Returned data should be freed with g_free() after usage.
  *
  * Returns: (transfer full): path string.
  *
  * Since: 0.1.0
  */
-/* FIXME: handle display name and real file name (maybe non-UTF8) issue */
 char* fm_path_to_str(FmPath* path)
 {
     gchar *ret;
@@ -764,6 +955,7 @@ char* fm_path_to_str(FmPath* path)
  * @path: a path
  *
  * Creates URI representation of @path.
+ * Returned data should be freed with g_free() after usage.
  *
  * Returns: path URI.
  *
@@ -787,49 +979,93 @@ char* fm_path_to_uri(FmPath* path)
 /**
  * fm_path_display_name
  * @path: a path
- * @human_readable: %TRUE to generate simple text
+ * @human_readable: ignored since 1.2.0
  *
- * Creates string representation of @path as displayable name.
+ * Creates string representation of @path as displayable UTF-8 string.
  * The conversion is the most probably unreversible so returned value
- * should be used only for displaying purposes.
+ * should be used only for displaying purposes. The displayable path
+ * may be reliable only if there was already #FmFileInfo retrieved for
+ * the @path, otherwise some estimation will be returned instead, be
+ * careful with this API.
+ * The main purpose for this API is to create displayable path for the
+ * path entry. Applications are encouraged to use g_file_get_parse_name()
+ * instead for any other purposes.
+ * Returned data should be freed with g_free() after usage.
+ *
+ * See also: fm_path_display_basename(), fm_path_to_str().
  *
  * Returns: (transfer full): path string.
  *
  * Since: 0.1.0
  */
 /* FIXME: maybe we can support different encoding for different mount points? */
-char* fm_path_display_name(FmPath* path, gboolean human_readable)
+static char *_fm_path_display_name(FmPath* path, gboolean *is_query)
 {
     char* disp;
-    if(human_readable)
+
+    if(G_LIKELY(path->parent))
     {
-        if(G_LIKELY(path->parent))
+        char* disp_parent;
+        char* disp_base;
+        if (G_UNLIKELY(*is_query)) /* already detected query */
+            goto _is_query;
+        else if (fm_path_is_native(path))
         {
-            char* disp_parent = fm_path_display_name(path->parent, TRUE);
-            char* disp_base = fm_path_display_basename(path);
-            disp = g_build_filename( disp_parent, disp_base, NULL);
-            g_free(disp_parent);
-            g_free(disp_base);
+            disp_parent = _fm_path_display_name(path->parent, is_query);
+            disp_base = fm_path_display_basename(path);
+        }
+        else if (strchr(path->name, '?') == NULL)
+        {
+            /* URI query not detected yet, check parent beforehand */
+            disp_parent = _fm_path_display_name(path->parent, is_query);
+            if (G_UNLIKELY(*is_query)) /* parent found that it is a query URI */
+                goto _is_query2;
+            else
+                disp_base = fm_path_display_basename(path);
         }
         else
-            disp = fm_path_display_basename(path);
+        {
+            /* it is not URI path but rather URI query, pass it "as is" */
+            *is_query = TRUE;
+_is_query:
+            disp_parent = _fm_path_display_name(path->parent, is_query);
+_is_query2:
+            disp = g_strjoin(NULL, disp_parent, path->name, NULL);
+            g_free(disp_parent);
+            return disp;
+        }
+        if (fm_path_is_native(path))
+            disp = g_build_filename( disp_parent, disp_base, NULL);
+        else
+            disp = g_build_path("/", disp_parent, disp_base, NULL);
+        g_free(disp_parent);
+        g_free(disp_base);
     }
     else
-    {
-        char* str = fm_path_to_str(path);
-        disp = g_filename_display_name(str);
-        g_free(str);
-    }
+        disp = g_uri_unescape_string(path->name, NULL);
     return disp;
 }
 
+char* fm_path_display_name(FmPath* path, gboolean human_readable)
+{
+    gboolean is_query = FALSE;
+    return _fm_path_display_name(path, &is_query);
+}
+
 /**
  * fm_path_display_basename
  * @path: a path
  *
- * Creates displayable basename of @path.
+ * Creates displayable basename of @path. If @path is a scheme root path
+ * (i.e. "/", "sftp://", etc.) then returns basename of @path. For other
+ * paths displayable basename may be reliable only if there was already
+ * #FmFileInfo retrieved for the @path, otherwise some estimation will be
+ * done instead, be careful with this API.
+ * Returned data should be freed with g_free() after usage.
  *
- * Returns: displayable basename of path.
+ * See also: fm_file_info_get_disp_name().
+ *
+ * Returns: (transfer full): displayable basename of path.
  *
  * Since: 0.1.0
  */
@@ -837,27 +1073,88 @@ char* fm_path_display_name(FmPath* path, gboolean human_readable)
 char* fm_path_display_basename(FmPath* path)
 {
     if(G_UNLIKELY(!path->parent)) /* root_path element */
+        return g_strdup(path->name);
+    G_LOCK(roots);
+    if (G_LIKELY(path->disp_name == BASENAME_AS_DISP_NAME))
     {
-        if( !fm_path_is_native(path) )
-        {
-            if(fm_path_is_trash_root(path))
-                return g_strdup(_("Trash Can"));
-            if(g_str_has_prefix(path->name, "computer:/"))
-                return g_strdup(_("My Computer"));
-            if(g_str_has_prefix(path->name, "network:/"))
-                return g_strdup(_("Network"));
-        }
+        G_UNLOCK(roots);
+        return g_strdup(path->name);
     }
+    if (path->disp_name)
+    {
+        char *name = g_strdup(path->disp_name);
+        G_UNLOCK(roots);
+        return name;
+    }
+    G_UNLOCK(roots);
     if(!fm_path_is_native(path))
         return g_uri_unescape_string(path->name, NULL);
     return g_filename_display_name(path->name);
 }
 
+void _fm_path_set_display_name(FmPath *path, const char *disp_name)
+{
+    if (!disp_name || !*disp_name)
+    {
+        char *_name = NULL;
+        if(!fm_path_is_native(path))
+            _name = g_uri_unescape_string(path->name, NULL);
+        else
+            _name = g_filename_display_name(path->name);
+        _fm_path_set_display_name(path, _name);
+        g_free(_name);
+        return;
+    }
+    G_LOCK(roots);
+    if (path->disp_name != BASENAME_AS_DISP_NAME)
+    {
+        /* check if it is set already */
+        if (g_strcmp0(disp_name, path->disp_name) == 0)
+        {
+            G_UNLOCK(roots);
+            return;
+        }
+        g_free(path->disp_name);
+    }
+    /* g_debug("set display name of %s to %s", path->name, disp_name); */
+    /* By default we use the real file base name for display.
+     * if the base name is not in UTF-8 encoding, we
+     * need to convert it to UTF-8 for display and save its
+     * UTF-8 version in fi->disp_name */
+    if (g_strcmp0(disp_name, path->name) == 0)
+        path->disp_name = BASENAME_AS_DISP_NAME;
+    else
+        path->disp_name = g_strdup(disp_name);
+    G_UNLOCK(roots);
+}
+
+/* use this to avoid change from another thread */
+static char *_display_name_static_keeper = NULL;
+
+/* this API is not thread capable! */
+const char *_fm_path_get_display_name(FmPath *path)
+{
+    G_LOCK(roots);
+    if (path->disp_name == BASENAME_AS_DISP_NAME)
+    {
+        G_UNLOCK(roots);
+        return path->name;
+    }
+    g_free(_display_name_static_keeper);
+    /* use this to keep disp_name after returning from lock because another
+       thread may change it at that time, although _display_name_static_keeper
+       isn't protected by lock so should be protected by general glib lock */
+    _display_name_static_keeper = g_strdup(path->disp_name);
+    G_UNLOCK(roots);
+    return _display_name_static_keeper;
+}
+
 /**
  * fm_path_to_gfile
  * @path: a path
  *
  * Creates #GFile representation of @path.
+ * Returned data should be freed with g_object_unref() after usage.
  *
  * Returns: (transfer full): a #GFile object.
  *
@@ -965,20 +1262,20 @@ void _fm_path_init()
 
     /* build path object for home dir */
     name = home_dir + 1; /* skip leading / */
-    parent = root_path;
+    parent = fm_path_ref(root_path);
     while((sep = strchr(name, '/')))
     {
         int len = (sep - name);
         if(len > 0)
         {
-            /* ref counting is not a problem here since this path component
-             * will exist till the termination of the program. So mem leak is ok. */
             tmp = _fm_path_new_internal(parent, name, len, FM_PATH_IS_LOCAL|FM_PATH_IS_NATIVE);
+            fm_path_unref(parent);
             parent = tmp;
         }
         name = sep + 1;
     }
     home_path = _fm_path_new_internal(parent, name, strlen(name), FM_PATH_IS_LOCAL|FM_PATH_IS_NATIVE);
+    fm_path_unref(parent);
 
     desktop_dir = g_get_user_special_dir(G_USER_DIRECTORY_DESKTOP);
     desktop_len = strlen(desktop_dir);
@@ -986,42 +1283,48 @@ void _fm_path_init()
         --desktop_len;
 
     /* build path object for desktop_path dir */
-    /* FIXME: can it be that desktop_dir is outside of home_dir ?
     if(G_UNLIKELY(strncmp(desktop_dir, home_dir, home_len)))
     {
+        /* can it be that desktop_dir is outside of home_dir ? */
         name = &desktop_dir[1];
-        parent = root_path;
-        while()
-        {
-        }
+        parent = fm_path_ref(root_path);
     }
     else
-    { */
-    name = desktop_dir + home_len + 1; /* skip home_path dir part / */
-    parent = home_path;
-    /* } */
+    {
+        name = desktop_dir + home_len; /* skip home_path dir part / */
+        while (*name == '/') name++; /* skip extra / if any */
+        if (*name == '\0')
+            name = "Desktop"; /* fallback: never use home_path as desktop_path */
+        parent = fm_path_ref(home_path); /* preserve ref */
+    }
     while((sep = strchr(name, '/')))
     {
         int len = (sep - name);
         if(len > 0)
         {
-            /* ref counting is not a problem here since this path component
-             * will exist till the termination of the program. So mem leak is ok. */
             tmp = _fm_path_new_internal(parent, name, len, FM_PATH_IS_LOCAL|FM_PATH_IS_NATIVE);
+            fm_path_unref(parent); /* parent is reffed by new child */
             parent = tmp;
         }
         name = sep + 1;
     }
     desktop_path = _fm_path_new_internal(parent, name, strlen(name), FM_PATH_IS_LOCAL|FM_PATH_IS_NATIVE);
+    fm_path_unref(parent); /* we used it already */
 
     /* build path object for trash can */
-    /* FIXME: currently there are problems with URIs. using trash:/ here will cause problems. */
     trash_root_path = _fm_path_new_internal(NULL, "trash:///", 9, FM_PATH_IS_TRASH|FM_PATH_IS_VIRTUAL|FM_PATH_IS_LOCAL);
+    _fm_path_set_display_name(trash_root_path, _("Trash Can"));
     apps_root_path = _fm_path_new_internal(NULL, "menu://applications/", 20, FM_PATH_IS_VIRTUAL|FM_PATH_IS_XDG_MENU);
 }
 
 void _fm_path_finalize(void)
 {
+    fm_path_unref(root_path);
+    fm_path_unref(home_path);
+    fm_path_unref(desktop_path);
+    fm_path_unref(trash_root_path);
+    fm_path_unref(apps_root_path);
+    root_path = home_path = desktop_path = trash_root_path = apps_root_path = NULL;
 }
 
 /* For used in hash tables */
@@ -1074,15 +1377,7 @@ guint fm_path_hash(FmPath* path)
  * use this. */
 gboolean fm_path_equal(FmPath* p1, FmPath* p2)
 {
-    if(p1 == p2)
-        return TRUE;
-    if(!p1) /* if p2 is also NULL then p1==p2 and that is handled above */
-        return FALSE;
-    if(!p2) /* case of p1==NULL handled above */
-        return FALSE;
-    if( strcmp(p1->name, p2->name) != 0 )
-        return FALSE;
-    return fm_path_equal( p1->parent, p2->parent);
+    return (p1 == p2);
 }
 
 /*
@@ -1218,10 +1513,16 @@ FmPathList* fm_path_list_new_from_uris(char* const* uris)
             FmPath* path;
             if(puri[0] == '/')
                 path = fm_path_new_for_path(puri);
-            else if(strstr(puri, "://"))
+            else
+            {
                 path = fm_path_new_for_uri(puri);
-            else /* it's not a valid path or URI */
-                continue;
+                if (path == root_path)
+                {
+                    /* it's not a valid path or URI */
+                    fm_path_unref(path);
+                    continue;
+                }
+            }
             fm_list_push_tail_noref((FmList*)pl, path);
         }
     }
@@ -1240,7 +1541,9 @@ FmPathList* fm_path_list_new_from_uris(char* const* uris)
  */
 FmPathList* fm_path_list_new_from_uri_list(const char* uri_list)
 {
-    char** uris = g_strsplit(uri_list, "\r\n", -1);
+    /* The line separator used by text/urilist is CRLF, not LF. However, some
+      existing implementations seems to use '\n'. Let's handle both. */
+    char** uris = g_strsplit_set(uri_list, "\r\n", -1);
     FmPathList* pl = fm_path_list_new_from_uris(uris);
     g_strfreev(uris);
     return pl;
