@@ -4,20 +4,21 @@
  *      Copyright 2009 - 2010 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
  *      Copyright 2012-2013 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of the Libfm library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -25,7 +26,7 @@
  * @short_description: Common utility functions used by libfm and libfm-gtk.
  * @title: Common Libfm utilities.
  *
- * @include: libfm/fm-utils.h
+ * @include: libfm/fm.h
  *
  * This scope contains common data parsing utilities.
  */
@@ -43,6 +44,7 @@
 #include <string.h>
 #include "fm-utils.h"
 #include "fm-file-info-job.h"
+#include "fm-config.h"
 
 #define BI_KiB  ((gdouble)1024.0)
 #define BI_MiB  ((gdouble)1024.0 * 1024.0)
@@ -68,32 +70,81 @@
  * Since: 0.1.0
  */
 char* fm_file_size_to_str( char* buf, size_t buf_size, goffset size, gboolean si_prefix )
+{
+    return fm_file_size_to_str2(buf, buf_size, size, si_prefix ? 'H' : 'h');
+}
+
+/**
+ * fm_file_size_to_str2
+ * @buf: pointer to array to make a string
+ * @buf_size: size of @buf
+ * @size: number to convert
+ * @size_units: type of units to convert
+ *
+ * Converts @size into text representation of form "21.4 kiB" for example.
+ * The @size_units defines which unit will be selected to convert (similar
+ * to ones that 'du' utility uses):
+ * - b : bytes
+ * - k : KiB (1024 bytes)
+ * - m : MiB (1048576 bytes)
+ * - g : GiB (1073741824 bytes)
+ * - K : kB (1000 bytes)
+ * - M : MB (1000000 bytes)
+ * - G : GB (1000000000 bytes)
+ * - h : adaptive in IEC units
+ * - H : adaptive in SI units
+ *
+ * Returns: @buf.
+ *
+ * Since: 0.1.0
+ */
+char *fm_file_size_to_str2(char *buf, size_t buf_size, goffset size, char size_units)
 {
     const char * unit;
     gdouble val;
 
-    if( si_prefix ) /* 1000 based SI units */
+    switch (size_units)
     {
+    case 'b': case 'B':
+        goto _b;
+    case 'k':
+        val = (gdouble)size;
+        goto _kib;
+    case 'K':
+        val = (gdouble)size;
+        goto _kb;
+    case 'm':
+        val = (gdouble)size;
+        goto _mib;
+    case 'M':
+        val = (gdouble)size;
+        goto _mb;
+    case 'g':
+        val = (gdouble)size;
+        goto _gib;
+    case 'G':
+        val = (gdouble)size;
+        goto _gb;
+    case 'H': /* 1000 based SI units */
+_si:
         if(size < (goffset)SI_KB)
-        {
-            snprintf(buf, buf_size,
-                     dngettext(GETTEXT_PACKAGE, "%u byte", "%u bytes", (gulong)size),
-                     (guint)size);
-            return buf;
-        }
+            goto _b;
         val = (gdouble)size;
         if(val < SI_MB)
         {
+_kb:
             val /= SI_KB;
-            unit = _("KB");
+            unit = _("kB");
         }
         else if(val < SI_GB)
         {
+_mb:
             val /= SI_MB;
             unit = _("MB");
         }
         else if(val < SI_TB)
         {
+_gb:
             val /= SI_GB;
             unit = _("GB");
         }
@@ -102,11 +153,14 @@ char* fm_file_size_to_str( char* buf, size_t buf_size, goffset size, gboolean si
             val /= SI_TB;
             unit = _("TB");
         }
-    }
-    else /* 1024-based binary prefix */
-    {
+        break;
+    default:
+        if(fm_config->si_unit)
+            goto _si;
+    case 'h': /* 1024-based binary prefix */
         if(size < (goffset)BI_KiB)
         {
+_b:
             snprintf(buf, buf_size,
                      dngettext(GETTEXT_PACKAGE, "%u byte", "%u bytes", (gulong)size),
                      (guint)size);
@@ -115,16 +169,19 @@ char* fm_file_size_to_str( char* buf, size_t buf_size, goffset size, gboolean si
         val = (gdouble)size;
         if(val < BI_MiB)
         {
+_kib:
             val /= BI_KiB;
             unit = _("KiB");
         }
         else if(val < BI_GiB)
         {
+_mib:
             val /= BI_MiB;
             unit = _("MiB");
         }
         else if(val < BI_TiB)
         {
+_gib:
             val /= BI_GiB;
             unit = _("GiB");
         }
@@ -326,7 +383,9 @@ char* fm_strdup_replace(char* str, char* old_str, char* new_str)
  * @user_data:  caller data to pass to callback
  *
  * This function parses line that contains some %&lt;char&gt; commands and does
- * substitutions on them using callbacks provided by caller.
+ * substitutions on them using callbacks provided by caller. Only options
+ * in @opts and %% will be expanded, all other %&lt;char&gt; sequences will
+ * be skipped.
  *
  * Returns: number of valid options found in @cmd.
  *
@@ -367,12 +426,10 @@ int fm_app_command_parse(const char* cmd, const FmAppCommandParseOption* opts,
                         ins = opt->callback(*c, user_data);
                         if(ins && *ins)
                             g_string_append(buf, ins);
-                        /* FIXME: add support for uri and escaping */
                     }
                     break;
                 }
             }
-            /* FIXME: should invalid options be passed 'as is' or ignored? */
         }
     }
     if(c != ptr)
@@ -388,7 +445,7 @@ int fm_app_command_parse(const char* cmd, const FmAppCommandParseOption* opts,
 static GMutex main_loop_run_mutex;
 static GCond main_loop_run_cond;
 #else
-G_LOCK_DEFINE(main_loop_run_mutex);
+G_LOCK_DEFINE_STATIC(main_loop_run_mutex);
 static GMutex *main_loop_run_mutex = NULL;
 static GCond *main_loop_run_cond = NULL;
 #endif
@@ -492,3 +549,61 @@ const char *fm_get_home_dir(void)
         homedir = g_get_home_dir();
     return homedir;
 }
+
+/**
+ * fm_uri_subpath_to_native_subpath
+ * @subpath: URI substring to convert
+ * @error: (out) (allow-none): location to set error
+ *
+ * Converts escaped URI subpath into file path in native encoding.
+ *
+ * Returns: (transfer full): transcoded path or %NULL in case of error.
+ *
+ * Since: 1.2.0
+ */
+char *fm_uri_subpath_to_native_subpath(const char *subpath, GError **error)
+{
+    char *unescaped, *native;
+
+    /* unescape path first */
+    unescaped = g_uri_unescape_string(subpath, "/");
+    if (unescaped == NULL)
+    {
+        g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_BAD_URI,
+                     _("The URI subpath '%s' contains invalid escaped characters"),
+                     subpath);
+        return NULL;
+    }
+    /* FIXME: try to gently convert characters, not return NULL */
+    native = g_filename_from_utf8(unescaped, -1, NULL, NULL, error);
+    g_free(unescaped);
+    return native;
+}
+
+/**
+ * fm_strcatv
+ * @strvp: pointer to strings list
+ * @astrv: (allow-none): list of strings to append
+ *
+ * Appends string list @astrv to @strvp replacing it.
+ *
+ * Since: 1.2.0
+ */
+void fm_strcatv(char ***strvp, char * const *astrv)
+{
+    guint i, len, len2;
+    char **new_strv;
+
+    if (astrv == NULL || *astrv == NULL) /* empty list */
+        return;
+    len = *strvp ? g_strv_length(*strvp) : 0;
+    len2 = g_strv_length((char **)astrv);
+    new_strv = g_new(char *, len + len2 + 1);
+    for (i = 0; i < len; i++)
+        new_strv[i] = (*strvp)[i];
+    for (i = 0; i < len2; i++)
+        new_strv[len+i] = g_strdup(astrv[i]);
+    new_strv[len+i] = NULL;
+    g_free(*strvp);
+    *strvp = new_strv;
+}
