@@ -1,7 +1,7 @@
-/* $Id$ */
 /*-
  * Copyright (c) 2004-2006  os-cillation e.K.
  * Copyright (c) 2002,2004  Anders Carlsson <andersca@gnu.org>
+ * Copyright 2024 Ingo Brückl
  *
  * Written by Benedikt Meurer <benny@xfce.org>.
  *
@@ -143,11 +143,12 @@ struct _ExoIconViewClass
 {
   GtkContainerClass __parent__;
 
+#if !GTK_CHECK_VERSION(3, 0, 0)
   /* virtual methods */
   void     (*set_scroll_adjustments)    (ExoIconView     *icon_view,
                                          GtkAdjustment   *hadjustment,
                                          GtkAdjustment   *vadjustment);
-
+#endif
   /* signals */
   void     (*item_activated)            (ExoIconView     *icon_view,
                                          GtkTreePath     *path);
@@ -230,6 +231,10 @@ guint                 exo_icon_view_get_single_click_timeout  (const ExoIconView
 void                  exo_icon_view_set_single_click_timeout  (ExoIconView              *icon_view,
                                                                guint                     single_click_timeout);
 
+gboolean              exo_icon_view_get_middle_click          (const ExoIconView        *icon_view);
+void                  exo_icon_view_set_middle_click          (ExoIconView              *icon_view,
+                                                               gboolean                  middle_click);
+
 void                  exo_icon_view_widget_to_icon_coords     (const ExoIconView        *icon_view,
                                                                gint                      wx,
                                                                gint                      wy,
@@ -313,7 +318,7 @@ gboolean              exo_icon_view_get_dest_item_at_pos      (ExoIconView
                                                                gint                      drag_y,
                                                                GtkTreePath             **path,
                                                                ExoIconViewDropPosition  *pos);
-GdkPixmap            *exo_icon_view_create_drag_icon          (ExoIconView              *icon_view,
+GdkPixbuf            *exo_icon_view_create_drag_icon          (ExoIconView              *icon_view,
                                                                GtkTreePath              *path);
 
 
