@@ -2,6 +2,7 @@
  *      fm-file-menu.c
  *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2013-2018 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -24,22 +25,26 @@
  * @short_description: Simple context menu for files.
  * @title: FmFileMenu
  *
- * @include: libfm/fm-file-menu.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmFileMenu can be used to create context menu on some file(s).
  *
  * The menu consists of items:
  * |[
  * Open
- * ------------------------
  * &lt;placeholder name='ph1'/&gt;
  * ------------------------
  * &lt;placeholder name='ph2'/&gt;
  * ------------------------
+ * AddBookmark
+ * &lt;placeholder name='SendToCategory'/&gt;
+ * ------------------------
  * Cut
  * Copy
  * Paste
  * Del
+ * CopyPath
+ * &lt;placeholder name='MoveCategory'/&gt;
  * ------------------------
  * Rename
  * ------------------------
@@ -50,18 +55,19 @@
  * You can modity the menu replacing placeholders. Note that internally
  * the menu constructor also puts some conditional elements into those
  * placeholders:
- * - ph1: 'UnTrash' item if files are in trash can
- * - ph2: 'OpenWith' list and submenu;
- * - ph3: custom (user defined) menu elements
+ * - ph2: 'OpenWith' list+selector (optionally in submenu 'OpenWithMenu');
+ * - SendToCategory: elements of category 'Send To':
  *  
  *        'Compress' if there is archiver defined;
  *  
  *        'Extract' if this is an archive
+ * - MoveCategory: 'Hide' or 'Unhide' if the attribute is changeable.
  *
- * Element 'Rename' is hidden if menu is created for more than one file.
- *
- * Elements 'Cut', 'Copy', 'Paste', and 'Del' are hidden for any virtual
- * place but trash can.
+ * Element 'AddBookmark' is visible only if menu is created for one directory.
+ * Element 'Rename' is hidden if menu is created for more than one file or
+ * if that file cannot be renamed.
+ * Element 'Paste' is visible only if menu is created for one directory
+ * and that directory is writable.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -72,38 +78,25 @@
 #include "../gtk-compat.h"
 
 #include "fm.h"
-#include "fm-config.h"
 
 #include "fm-file-menu.h"
-#include "fm-path.h"
 
 #include "fm-clipboard.h"
 #include "fm-file-properties.h"
-#include "fm-utils.h"
 #include "fm-gtk-utils.h"
 #include "fm-app-chooser-dlg.h"
-#include "fm-archiver.h"
-#include "fm-app-info.h"
 #include "fm-gtk-file-launcher.h"
 
-#ifdef HAVE_ACTIONS
-#include "fm-actions.h"
-#endif
-
-#include "gtk-compat.h"
-
 struct _FmFileMenu
 {
     FmFileInfoList* file_infos;
     gboolean same_type : 1;
-    gboolean same_fs : 1;
-    gboolean all_virtual : 1;
-    gboolean all_trash : 1;
-    gboolean auto_destroy : 1; // private
+    gboolean can_trash : 1;
+    //gboolean disable_archiving : 1;
+    gboolean shift_pressed : 1;
     GtkUIManager* ui;
     GtkActionGroup* act_grp;
     GtkMenu* menu;
-    GtkWindow* parent;
 
     FmLaunchFolderFunc folder_func;
     gpointer folder_func_data;
@@ -118,7 +111,10 @@ static void on_cut(GtkAction* action, gpointer user_data);
 static void on_copy(GtkAction* action, gpointer user_data);
 static void on_paste(GtkAction* action, gpointer user_data);
 static void on_delete(GtkAction* action, gpointer user_data);
-static void on_untrash(GtkAction* action, gpointer user_data);
+static void on_copy_path(GtkAction* action, gpointer user_data);
+static void on_hide(GtkAction* action, FmFileMenu* menu);
+static void on_unhide(GtkAction* action, FmFileMenu* menu);
+static void on_add_bookmark(GtkAction* action, FmFileMenu* menu);
 static void on_rename(GtkAction* action, gpointer user_data);
 static void on_compress(GtkAction* action, gpointer user_data);
 static void on_extract_here(GtkAction* action, gpointer user_data);
@@ -128,22 +124,21 @@ static void on_prop(GtkAction* action, gpointer user_data);
 const char base_menu_xml[]=
 "<popup>"
   "<menuitem action='Open'/>"
-  "<separator/>"
   "<placeholder name='ph1'/>"
   "<separator/>"
   "<placeholder name='ph2'/>"
   "<separator/>"
+  "<menuitem action='AddBookmark'/>"
+  "<placeholder name='SendToCategory'/>"
+  "<separator/>"
   "<menuitem action='Cut'/>"
   "<menuitem action='Copy'/>"
   "<menuitem action='Paste'/>"
   "<menuitem action='Del'/>"
+  "<menuitem action='CopyPath'/>"
+  "<placeholder name='MoveCategory'/>"
   "<separator/>"
   "<menuitem action='Rename'/>"
-/* TODO: implement symlink creation and "send to".
-  "<menuitem action='Link'/>"
-  "<menu action='SendTo'>"
-  "</menu>"
-*/
   "<separator/>"
   "<placeholder name='ph3'/>"
   "<separator/>"
@@ -155,20 +150,31 @@ GtkActionEntry base_menu_actions[]=
 {
     {"Open", GTK_STOCK_OPEN, N_("_Open"), NULL, NULL, G_CALLBACK(on_open)},
     {"OpenWith", NULL, N_("Open _With..."), NULL, NULL, G_CALLBACK(on_open_with)},
-    {"OpenWithMenu", NULL, N_("Open _With..."), NULL, NULL, NULL},
+    {"OpenWithMenu", NULL, N_("Open _With"), NULL, NULL, NULL},
     {"Cut", GTK_STOCK_CUT, NULL, NULL, NULL, G_CALLBACK(on_cut)},
     {"Copy", GTK_STOCK_COPY, NULL, NULL, NULL, G_CALLBACK(on_copy)},
     {"Paste", GTK_STOCK_PASTE, NULL, NULL, NULL, G_CALLBACK(on_paste)},
     {"Del", GTK_STOCK_DELETE, NULL, NULL, NULL, G_CALLBACK(on_delete)},
-    {"Rename", NULL, N_("_Rename"), NULL, NULL, G_CALLBACK(on_rename)},
-    {"Link", NULL, N_("Create _Symlink"), NULL, NULL, NULL},
-    {"SendTo", NULL, N_("Se_nd To"), NULL, NULL, NULL},
+    {"CopyPath", NULL, N_("Copy Pa_th(s)"), NULL, NULL, G_CALLBACK(on_copy_path)},
+    {"Hide", NULL, N_("H_ide"), NULL, NULL, G_CALLBACK(on_hide)},
+    {"Unhide", NULL, N_("Unh_ide"), NULL, NULL, G_CALLBACK(on_unhide)},
+    {"AddBookmark", GTK_STOCK_ADD, N_("_Add to Bookmarks"), NULL, NULL, G_CALLBACK(on_add_bookmark)},
+    {"Rename", NULL, N_("_Rename..."), NULL, NULL, G_CALLBACK(on_rename)},
     {"Compress", NULL, N_("Co_mpress..."), NULL, NULL, G_CALLBACK(on_compress)},
     {"Extract", NULL, N_("Extract _Here"), NULL, NULL, G_CALLBACK(on_extract_here)},
-    {"Extract2", NULL, N_("E_xtract To..."), NULL, NULL, G_CALLBACK(on_extract_to)},
+    {"ExtractTo", NULL, N_("E_xtract To..."), NULL, NULL, G_CALLBACK(on_extract_to)},
     {"Prop", GTK_STOCK_PROPERTIES, N_("Prop_erties"), NULL, NULL, G_CALLBACK(on_prop)}
 };
 
+/* plugins for MIME type */
+typedef struct {
+    FmMimeType *mime_type;
+    FmFileMenuMimeAddonInit cb; /* callbacks */
+} FmFileMenuMimeExt;
+
+static GList *extensions = NULL; /* elements are FmFileMenuMimeExt */
+
+
 /**
  * fm_file_menu_destroy
  * @menu: a menu
@@ -179,11 +185,13 @@ GtkActionEntry base_menu_actions[]=
  */
 void fm_file_menu_destroy(FmFileMenu* menu)
 {
-    if(menu->parent)
-        g_object_unref(menu->parent);
+    GObject *win = G_OBJECT(gtk_menu_get_attach_widget(menu->menu));
 
-    if(menu->menu)
-        gtk_widget_destroy(GTK_WIDGET(menu->menu));
+    if (win)
+        g_object_weak_unref(win, (GWeakNotify)gtk_menu_detach, menu->menu);
+
+    gtk_menu_detach(menu->menu);
+    gtk_widget_destroy(GTK_WIDGET(menu->menu));
 
     if(menu->file_infos)
         fm_file_info_list_unref(menu->file_infos);
@@ -219,96 +227,54 @@ FmFileMenu* fm_file_menu_new_for_file(GtkWindow* parent, FmFileInfo* fi, FmPath*
     return menu;
 }
 
-#ifdef HAVE_ACTIONS
-static void on_custom_action(GtkAction* act, FmFileMenu* data)
+static gboolean on_key_pressed(GtkWidget *widget, GdkEventKey *event, FmFileMenu *data)
 {
-    FmFileActionItem* item = FM_FILE_ACTION_ITEM(g_object_get_qdata(G_OBJECT(act), fm_qdata_id));
-    GdkAppLaunchContext* ctx = gdk_display_get_app_launch_context(gdk_display_get_default());
-    GList* files = fm_file_info_list_peek_head_link(data->file_infos);
-    char* output = NULL;
-    gdk_app_launch_context_set_screen(ctx, gtk_widget_get_screen(GTK_WIDGET(data->menu)));
-    gdk_app_launch_context_set_timestamp(ctx, gtk_get_current_event_time());
+    GtkAction *act;
+    GdkModifierType mask = 0;
 
-    /* g_debug("item: %s is activated, id:%s", fm_file_action_item_get_name(item),
-        fm_file_action_item_get_id(item)); */
-    fm_file_action_item_launch(item, G_APP_LAUNCH_CONTEXT(ctx), files, &output);
-    if(output)
+    gdk_window_get_device_position (gtk_widget_get_window(widget),
+                                    gdk_device_manager_get_client_pointer(
+                                        gdk_display_get_device_manager(
+                                            gdk_screen_get_display(
+                                                gtk_widget_get_screen(widget)))),
+                                    NULL, NULL, &mask);
+    if (mask & GDK_SHIFT_MASK)
     {
-        fm_show_error(NULL, "output", output);
-        g_free(output);
-    }
-    g_object_unref(ctx);
-}
-
-static void add_custom_action_item(FmFileMenu* data, GString* xml, FmFileActionItem* item)
-{
-    GtkAction* act;
-    if(!item) /* separator */
-    {
-        g_string_append(xml, "<separator/>");
-        return;
-    }
-
-    if(fm_file_action_item_is_action(item))
-    {
-        if(!(fm_file_action_item_get_target(item) & FM_FILE_ACTION_TARGET_CONTEXT))
-            return;
-    }
-
-    act = gtk_action_new(fm_file_action_item_get_id(item),
-                         fm_file_action_item_get_name(item),
-                         fm_file_action_item_get_desc(item),
-                         NULL);
-
-    if(fm_file_action_item_is_action(item))
-        g_signal_connect(act, "activate", G_CALLBACK(on_custom_action), data);
-
-    gtk_action_set_icon_name(act, fm_file_action_item_get_icon(item));
-    gtk_action_group_add_action(data->act_grp, act);
-    /* associate the app info object with the action */
-    g_object_set_qdata_full(G_OBJECT(act), fm_qdata_id,
-                            fm_file_action_item_ref(item),
-                            fm_file_action_item_unref);
-    if(fm_file_action_item_is_menu(item))
-    {
-        GList* subitems = fm_file_action_item_get_sub_items(item);
-        GList* l;
-        g_string_append_printf(xml, "<menu action='%s'>",
-                               fm_file_action_item_get_id(item));
-        for(l=subitems; l; l=l->next)
+        if (!data->shift_pressed)
         {
-            FmFileActionItem* subitem = FM_FILE_ACTION_ITEM(l->data);
-            add_custom_action_item(data, xml, subitem);
+            data->shift_pressed = TRUE;
+            act = gtk_ui_manager_get_action(data->ui, "/popup/Del");
+            gtk_action_set_label(act, NULL);
+            gtk_action_set_stock_id(act, GTK_STOCK_REMOVE);
         }
-        g_string_append(xml, "</menu>");
-    }
-    else
-    {
-        g_string_append_printf(xml, "<menuitem action='%s'/>",
-                               fm_file_action_item_get_id(item));
     }
+    return FALSE;
 }
 
-static void fm_file_menu_add_custom_actions(FmFileMenu* data, GString* xml, FmFileInfoList* files)
+static gboolean on_key_released(GtkWidget *widget, GdkEventKey *event, FmFileMenu *data)
 {
-    GList* files_list = fm_file_info_list_peek_head_link(files);
-    GList* items = fm_get_actions_for_files(files_list);
+    GtkAction *act;
+    GdkModifierType mask = 0;
 
-    if(items)
+    gdk_window_get_device_position (gtk_widget_get_window(widget),
+                                    gdk_device_manager_get_client_pointer(
+                                        gdk_display_get_device_manager(
+                                            gdk_screen_get_display(
+                                                gtk_widget_get_screen(widget)))),
+                                    NULL, NULL, &mask);
+    if (!(mask & GDK_SHIFT_MASK))
     {
-        g_string_append(xml, "<popup><placeholder name='ph3'>");
-        GList* l;
-        for(l=items; l; l=l->next)
+        if (data->shift_pressed)
         {
-            FmFileActionItem* item = FM_FILE_ACTION_ITEM(l->data);
-            add_custom_action_item(data, xml, item);
+            data->shift_pressed = FALSE;
+            act = gtk_ui_manager_get_action(data->ui, "/popup/Del");
+            /* Note to translators: Trash in not noun but verb here */
+            gtk_action_set_label(act, _("Move to _Trash"));
+            gtk_action_set_stock_id(act, GTK_STOCK_DELETE);
         }
-        g_string_append(xml, "</placeholder></popup>");
     }
-    g_list_foreach(items, (GFunc)fm_file_action_item_unref, NULL);
-    g_list_free(items);
+    return FALSE;
 }
-#endif /* HAVE_ACTIONS */
 
 /**
  * fm_file_menu_new_for_files
@@ -328,27 +294,17 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
     GtkUIManager* ui;
     GtkActionGroup* act_grp;
     GtkAction* act;
-    FmFileInfo* fi = fm_file_info_list_peek_head(files);
+    FmFileInfo* fi;
     FmFileMenu* data = g_slice_new0(FmFileMenu);
     GString* xml;
     GList* mime_types = NULL;
     GList* l;
     GList* apps = NULL;
-    FmPath* path = fm_file_info_get_path(fi);
-
+    gboolean all_native = TRUE;
     unsigned items_num = fm_file_info_list_get_length(files);
 
-    data->parent = g_object_ref(parent); /* FIXME: is this really needed? */
-    /* FIXME: should we connect to "destroy" signal of parent and set data->parent to NULL when
-     * it's detroyed? */
     data->file_infos = fm_file_info_list_ref(files);
 
-    /* check if the files are on the same filesystem */
-    data->same_fs = fm_file_info_list_is_same_fs(files);
-
-    data->all_virtual = data->same_fs && !fm_path_is_native(path);
-    data->all_trash = data->same_fs && fm_path_is_trash(path);
-
     /* create list of mime types */
     for(l = fm_file_info_list_peek_head_link(files); l; l = l->next)
     {
@@ -356,8 +312,18 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
         GList* l2;
 
         fi = l->data;
+        if (!fm_file_info_is_native(fi))
+            all_native = FALSE;
+        else if (all_native && fm_file_info_is_shortcut(fi))
+        {
+            /* this takes some time but it's inevitable */
+            FmPath *tp = fm_path_new_for_str(fm_file_info_get_target(fi));
+            if (!fm_path_is_native(tp))
+                all_native = FALSE;
+            fm_path_unref(tp);
+        }
         mime_type = fm_file_info_get_mime_type(fi);
-        if(mime_type == NULL || !fm_file_info_is_native(fi))
+        if(mime_type == NULL)
             continue;
         for(l2 = mime_types; l2; l2 = l2->next)
             if(l2->data == mime_type)
@@ -366,6 +332,7 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
             continue;
         mime_types = g_list_prepend(mime_types, fm_mime_type_ref(mime_type));
     }
+    fi = fm_file_info_list_peek_head(files); /* we'll test it below */
     /* create apps list */
     if(mime_types)
     {
@@ -395,12 +362,8 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
             g_list_foreach(apps2, (GFunc)g_object_unref, NULL);
             g_list_free(apps2);
         }
-        g_list_foreach(mime_types, (GFunc)fm_mime_type_unref, NULL);
-        g_list_free(mime_types);
-        mime_types = NULL;
     }
 
-    data->auto_destroy = auto_destroy;
     data->ui = ui = gtk_ui_manager_new();
     data->act_grp = act_grp = gtk_action_group_new("Popup");
     gtk_action_group_set_translation_domain(act_grp, GETTEXT_PACKAGE);
@@ -415,155 +378,133 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
     xml = g_string_new("<popup><placeholder name='ph2'>");
     if(apps) /* add specific menu items for those files */
     {
-            gboolean use_sub = g_list_length(apps) > 5;
-            if(use_sub)
-                g_string_append(xml, "<menu action='OpenWithMenu'>");
+        gboolean use_sub = g_list_length(apps) > 5;
+        gboolean found_app = FALSE;
+        if(use_sub)
+            g_string_append(xml, "<menu action='OpenWithMenu'>");
 
-            for(l=apps;l;l=l->next)
+        for(l=apps;l;l=l->next)
+        {
+            GAppInfo* app = l->data;
+
+            /*g_debug("app %s, executable %s, command %s\n",
+                g_app_info_get_name(app),
+                g_app_info_get_executable(app),
+                g_app_info_get_commandline(app));*/
+
+            gchar * program_path = g_find_program_in_path(g_app_info_get_executable(app));
+            if (!program_path)
+                goto _next_app;
+            g_free(program_path);
+            if (!all_native && !g_app_info_supports_uris(app))
             {
-                GAppInfo* app = l->data;
-
-                /*g_debug("app %s, executable %s, command %s\n",
-                    g_app_info_get_name(app),
-                    g_app_info_get_executable(app),
-                    g_app_info_get_commandline(app));*/
-
-                gchar * program_path = g_find_program_in_path(g_app_info_get_executable(app));
-                if (!program_path)
-                    continue;
-                g_free(program_path);
-
-                act = gtk_action_new(g_app_info_get_id(app),
-                                     g_app_info_get_name(app),
-                                     g_app_info_get_description(app),
-                                     NULL);
-                g_signal_connect(act, "activate", G_CALLBACK(on_open_with_app), data);
-                gtk_action_set_gicon(act, g_app_info_get_icon(app));
-                gtk_action_group_add_action(act_grp, act);
-                /* associate the app info object with the action */
-                g_object_set_qdata_full(G_OBJECT(act), fm_qdata_id, app, g_object_unref);
-                g_string_append_printf(xml, "<menuitem action='%s'/>", g_app_info_get_id(app));
+_next_app:
+                g_object_unref(app);
+                continue;
             }
+            found_app = TRUE;
 
-            g_list_free(apps); /* don't unref GAppInfos now */
-            if(use_sub)
-            {
-                g_string_append(xml,
-                                "<separator/>"
-                                "<menuitem action='OpenWith'/>"
-                                "</menu>");
-            }
-            else
-                g_string_append(xml, "<menuitem action='OpenWith'/>");
+            act = gtk_action_new(g_app_info_get_id(app),
+                                 g_app_info_get_name(app),
+                                 g_app_info_get_description(app),
+                                 NULL);
+            g_signal_connect(act, "activate", G_CALLBACK(on_open_with_app), data);
+            gtk_action_set_gicon(act, g_app_info_get_icon(app));
+            gtk_action_group_add_action(act_grp, act);
+            g_object_unref(act);
+            /* associate the app info object with the action */
+            g_object_set_qdata_full(G_OBJECT(act), fm_qdata_id, app, g_object_unref);
+            g_string_append_printf(xml, "<menuitem action='%s'/>", g_app_info_get_id(app));
+        }
+
+        g_list_free(apps); /* don't unref GAppInfos now */
+        if (!found_app)
+            goto _disable_open;
+        if(use_sub)
+        {
+            g_string_append(xml, "<separator/>"
+                                 "<menuitem action='OpenWith'/>"
+                                 "</menu>");
+        }
+        else
+            g_string_append(xml, "<menuitem action='OpenWith'/>");
     }
     else
     {
+_disable_open:
         act = gtk_ui_manager_get_action(ui, "/popup/Open");
         gtk_action_set_visible(act, FALSE);
         g_string_append(xml, "<menuitem action='OpenWith'/>");
     }
     g_string_append(xml, "</placeholder></popup>");
-
-#ifdef HAVE_ACTIONS
-    /* add custom file actions */
-    fm_file_menu_add_custom_actions(data, xml, files);
-#endif
-
-    /* archiver integration */
-    if(!data->all_virtual)
+    if (items_num == 1 && fm_file_info_can_set_hidden(fi))
     {
-        g_string_append(xml, "<popup><placeholder name='ph3'>");
-        if(data->same_type)
-        {
-            FmArchiver* archiver = fm_archiver_get_default();
-            if(archiver)
-            {
-                FmMimeType* mime_type;
-                fi = fm_file_info_list_peek_head(files);
-                mime_type = fm_file_info_get_mime_type(fi);
-                if(fm_archiver_is_mime_type_supported(archiver, fm_mime_type_get_type(mime_type)))
-                {
-                    if(data->cwd && archiver->extract_to_cmd)
-                        g_string_append(xml, "<menuitem action='Extract'/>");
-                    if(archiver->extract_cmd)
-                        g_string_append(xml, "<menuitem action='Extract2'/>");
-                }
-                else
-                    g_string_append(xml, "<menuitem action='Compress'/>");
-            }
-        }
+        g_string_append(xml, "<popup><placeholder name='MoveCategory'>");
+        if (fm_file_info_is_hidden(fi))
+            g_string_append(xml, "<menuitem action='Unhide'/>");
         else
-            g_string_append(xml, "<menuitem action='Compress'/>");
+            g_string_append(xml, "<menuitem action='Hide'/>");
         g_string_append(xml, "</placeholder></popup>");
     }
-
-    /* Special handling for some virtual filesystems */
-    /* FIXME: it should be done on per-scheme basis */
-    g_string_append(xml, "<popup><placeholder name='ph1'>");
-    if(data->all_virtual)
+    if (data->same_type)
     {
-        /* if all of the files are in trash */
-        if(data->all_trash)
+        CHECK_MODULES();
+        /* run mime-specific extensions */
+        for (l = extensions; l; l = l->next)
         {
-            gboolean can_restore = TRUE;
-            GList* l;
-            /* only immediate children of trash:/// can be restored. */
-            for(l = fm_file_info_list_peek_head_link(files);l;l=l->next)
-            {
-                FmPath* trash_path = fm_file_info_get_path(FM_FILE_INFO(l->data));
-                if(!fm_path_get_parent(trash_path) ||
-                   !fm_path_is_trash_root(fm_path_get_parent(trash_path)))
-                {
-                    can_restore = FALSE;
-                    break;
-                }
-            }
-
-            if(can_restore)
-            {
-                act = gtk_action_new("UnTrash",
-                                    _("_Restore"),
-                                    _("Restore trashed files to original paths"),
-                            NULL);
-                g_signal_connect(act, "activate", G_CALLBACK(on_untrash), data);
-                gtk_action_group_add_action(act_grp, act);
-                g_string_append(xml, "<menuitem action='UnTrash'/>");
-            }
-
-            act = gtk_ui_manager_get_action(ui, "/popup/Open");
-            gtk_action_set_visible(act, FALSE);
+            register FmFileMenuMimeExt *ext = l->data;
+            if (ext->mime_type == mime_types->data) /* the same mime type */
+                if (ext->cb.update_file_menu_for_mime_type != NULL)
+                    ext->cb.update_file_menu_for_mime_type(parent, ui, xml,
+                                                           act_grp, data, files,
+                                                           (items_num == 1));
         }
-        else
-        {
-            /* do not provide these items for other virtual files */
-            /* FIXME: this is invalid way to do it, many virtual paths still
-               support getting files content and some can deletion as well.
-               Cut/Paste/Rename/Del should be disabled only if FS is R/O and
-               Copy only for very few FS such as trash:// or menu://
-               Since query if FS is R/O is async operation we have to get
-               such info from FmFolder therefore do this in fm-folder-view.c */
-            act = gtk_ui_manager_get_action(ui, "/popup/Cut");
-            gtk_action_set_visible(act, FALSE);
-            act = gtk_ui_manager_get_action(ui, "/popup/Copy");
-            gtk_action_set_visible(act, FALSE);
-            act = gtk_ui_manager_get_action(ui, "/popup/Paste");
-            gtk_action_set_visible(act, FALSE);
-            act = gtk_ui_manager_get_action(ui, "/popup/Del");
-            gtk_action_set_visible(act, FALSE);
-        }
-        act = gtk_ui_manager_get_action(ui, "/popup/Rename");
-        gtk_action_set_visible(act, FALSE);
     }
-    g_string_append(xml, "</placeholder></popup>");
+
+    /* archiver integration */
+    if (all_native)
+    {
+        FmArchiver* archiver = fm_archiver_get_default();
+        if(archiver)
+        {
+            g_string_append(xml, "<popup><placeholder name='SendToCategory'>");
+            for (l = mime_types; l; l = l->next)
+                if (!fm_archiver_is_mime_type_supported(archiver, fm_mime_type_get_type(l->data)))
+                    break;
+            if (mime_types && l == NULL) /* all are archives */
+            {
+                if(data->cwd && archiver->extract_to_cmd)
+                    g_string_append(xml, "<menuitem action='ExtractTo'/>");
+                if(archiver->extract_cmd)
+                    g_string_append(xml, "<menuitem action='Extract'/>");
+            }
+            else
+                g_string_append(xml, "<menuitem action='Compress'/>");
+            g_string_append(xml, "</placeholder></popup>");
+        }
+    }
+    g_list_foreach(mime_types, (GFunc)fm_mime_type_unref, NULL);
+    g_list_free(mime_types);
 
     /* shadow 'Paste' if clipboard is empty and unshadow if not */
     act = gtk_ui_manager_get_action(ui, "/popup/Paste");
-    if(items_num != 1 || !fm_file_info_is_dir(fm_file_info_list_peek_head(files)))
+    if(items_num != 1 || !fm_file_info_is_dir(fi))
+    {
+        gtk_action_set_visible(act, FALSE);
+        act = gtk_ui_manager_get_action(ui, "/popup/AddBookmark");
+        gtk_action_set_visible(act, FALSE);
+    }
+    else if (!fm_file_info_is_writable_directory(fi))
+        /* it is still directory but file system is R/O */
         gtk_action_set_visible(act, FALSE);
     else
         gtk_action_set_sensitive(act, fm_clipboard_have_files(GTK_WIDGET(parent)));
 
-    if (items_num != 1)
+    if (items_num != 1 || !fm_file_info_can_set_name(fi) ||
+        /* unfortunately GIO doesn't support changing the .desktop files
+           but treats them as usual files, we can rename only file names,
+           not display names so we have no choice but disable this option */
+        fm_file_info_is_shortcut(fi) || fm_file_info_is_desktop_entry(fi))
     {
         act = gtk_ui_manager_get_action(ui, "/popup/Rename");
         gtk_action_set_visible(act, FALSE);
@@ -571,6 +512,27 @@ FmFileMenu* fm_file_menu_new_for_files(GtkWindow* parent, FmFileInfoList* files,
 
     gtk_ui_manager_add_ui_from_string(ui, xml->str, xml->len, NULL);
 
+    data->menu = GTK_MENU(gtk_ui_manager_get_widget(data->ui, "/popup"));
+    gtk_menu_attach_to_widget(data->menu, GTK_WIDGET(parent), NULL);
+    g_object_weak_ref(G_OBJECT(parent), (GWeakNotify)gtk_menu_detach, data->menu);
+    fm_widget_menu_fix_tooltips(data->menu);
+
+    if(auto_destroy)
+    {
+        g_signal_connect_swapped(data->menu, "selection-done",
+                                 G_CALLBACK(fm_file_menu_destroy), data);
+    }
+
+    if (fm_config->use_trash && all_native)
+    {
+        data->can_trash = TRUE;
+        /* change menu item text&icon when Shift is pressed */
+        g_signal_connect(data->menu, "key-press-event", G_CALLBACK(on_key_pressed), data);
+        g_signal_connect(data->menu, "key-release-event", G_CALLBACK(on_key_released), data);
+        act = gtk_ui_manager_get_action(ui, "/popup/Del");
+        gtk_action_set_label(act, _("Move to _Trash"));
+    }
+
     g_string_free(xml, TRUE);
     return data;
 }
@@ -624,11 +586,11 @@ FmFileInfoList* fm_file_menu_get_file_info_list(FmFileMenu* menu)
 }
 
 /**
- * fm_file_menu_get_ui
+ * fm_file_menu_get_menu
  * @menu: a menu
  *
- * Builds the menu with GtkUIManager. Returned data are owned by @menu
- * and should be not freed by caller.
+ * Retrieves #GtkMenu widget built with GtkUIManager. Returned data are
+ * owned by @menu and should be not freed by caller.
  *
  * Returns: (transfer none): created #GtkMenu widget.
  *
@@ -636,25 +598,31 @@ FmFileInfoList* fm_file_menu_get_file_info_list(FmFileMenu* menu)
  */
 GtkMenu* fm_file_menu_get_menu(FmFileMenu* menu)
 {
-    if( ! menu->menu )
-    {
-        menu->menu = GTK_MENU(gtk_ui_manager_get_widget(menu->ui, "/popup"));
-        gtk_menu_attach_to_widget(menu->menu, GTK_WIDGET(menu->parent), NULL);
-
-        if(menu->auto_destroy)
-        {
-            g_signal_connect_swapped(menu->menu, "selection-done",
-                                     G_CALLBACK(fm_file_menu_destroy), menu);
-        }
-    }
     return menu->menu;
 }
 
+/**
+ * fm_file_menu_get_cwd
+ * @menu: a menu
+ *
+ * Retrieves working directory @menu was created for. Returned data are owned
+ * by @menu and should be not freed by caller.
+ *
+ * Returns: (transfer none): working directory path.
+ *
+ * Since: 1.3.0
+ */
+FmPath* fm_file_menu_get_cwd(FmFileMenu* menu)
+{
+    return menu->cwd;
+}
+
 static void on_open(GtkAction* action, gpointer user_data)
 {
     FmFileMenu* data = (FmFileMenu*)user_data;
     GList* l = fm_file_info_list_peek_head_link(data->file_infos);
-    fm_launch_files_simple(data->parent, NULL, l, data->folder_func, data->folder_func_data);
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(data->menu));
+    fm_launch_files_simple(window, NULL, l, data->folder_func, data->folder_func_data);
 }
 
 static void open_with_app(FmFileMenu* data, GAppInfo* app)
@@ -667,8 +635,16 @@ static void open_with_app(FmFileMenu* data, GAppInfo* app)
     for(i=0; l; ++i, l=l->next)
     {
         FmFileInfo* fi = FM_FILE_INFO(l->data);
-        FmPath* path = fm_file_info_get_path(fi);
-        char* uri = fm_path_to_uri(path);
+        FmPath *path;
+        char *uri;
+        /* handle shortcuts here */
+        if (fm_file_info_is_shortcut(fi))
+            uri = g_strdup(fm_file_info_get_target(fi));
+        else
+        {
+            path = fm_file_info_get_path(fi);
+            uri = fm_path_to_uri(path);
+        }
         uris = g_list_prepend(uris, uri);
     }
     uris = g_list_reverse(uris);
@@ -701,20 +677,18 @@ static void on_open_with(GtkAction* action, gpointer user_data)
     FmFileInfo* fi = fm_file_info_list_peek_head(files);
     FmMimeType* mime_type;
     GAppInfo* app;
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(data->menu));
 
     if(data->same_type)
         mime_type = fm_file_info_get_mime_type(fi);
     else
         mime_type = NULL;
 
-    app = fm_choose_app_for_mime_type(data->parent, mime_type, TRUE);
+    app = fm_choose_app_for_mime_type(window, mime_type, TRUE);
 
     if(app)
     {
         open_with_app(data, app);
-        /* add the app to apps that support this file type. */
-        if(mime_type)
-            g_app_info_add_supports_type(app, fm_mime_type_get_type(mime_type), NULL);
         g_object_unref(app);
     }
 }
@@ -724,7 +698,7 @@ static void on_cut(GtkAction* action, gpointer user_data)
     FmFileMenu* data = (FmFileMenu*)user_data;
     FmPathList* files;
     files = fm_path_list_new_from_file_info_list(data->file_infos);
-    fm_clipboard_cut_files(GTK_WIDGET(data->parent), files);
+    fm_clipboard_cut_files(gtk_menu_get_attach_widget(data->menu), files);
     fm_path_list_unref(files);
 }
 
@@ -733,7 +707,7 @@ static void on_copy(GtkAction* action, gpointer user_data)
     FmFileMenu* data = (FmFileMenu*)user_data;
     FmPathList* files;
     files = fm_path_list_new_from_file_info_list(data->file_infos);
-    fm_clipboard_copy_files(GTK_WIDGET(data->parent), files);
+    fm_clipboard_copy_files(gtk_menu_get_attach_widget(data->menu), files);
     fm_path_list_unref(files);
 }
 
@@ -743,7 +717,8 @@ static void on_paste(GtkAction* action, gpointer user_data)
     FmFileInfo* fi = fm_file_info_list_peek_head(data->file_infos);
     if (fi)
     {
-        fm_clipboard_paste_files(GTK_WIDGET(data->parent), fm_file_info_get_path(fi));
+        fm_clipboard_paste_files(gtk_menu_get_attach_widget(data->menu),
+                                 fm_file_info_get_path(fi));
     }
 }
 
@@ -751,36 +726,83 @@ static void on_delete(GtkAction* action, gpointer user_data)
 {
     FmFileMenu* data = (FmFileMenu*)user_data;
     FmPathList* files;
-    GdkModifierType mask = 0;
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(data->menu));
+
     files = fm_path_list_new_from_file_info_list(data->file_infos);
     /* Fix for #3436283: accept Shift to delete instead of trash */
-    /* FIXME: change menu item text&icon when Shift is pressed */
-    gdk_window_get_device_position (gtk_widget_get_window(GTK_WIDGET(data->menu)),
-                                    gtk_get_current_event_device(),
-                                    NULL, NULL, &mask);
-    if(mask & GDK_SHIFT_MASK)
-        fm_delete_files(data->parent, files);
+    if (!data->can_trash || data->shift_pressed)
+        fm_delete_files(window, files);
     else
-        fm_trash_or_delete_files(data->parent, files);
+        fm_trash_or_delete_files(window, files);
     fm_path_list_unref(files);
 }
 
-static void on_untrash(GtkAction* action, gpointer user_data)
+static void on_copy_path(GtkAction* action, gpointer user_data)
 {
     FmFileMenu* data = (FmFileMenu*)user_data;
-    FmPathList* files;
-    files = fm_path_list_new_from_file_info_list(data->file_infos);
-    fm_untrash_files(data->parent, files);
-    fm_path_list_unref(files);
+    GList *fl;
+    GtkWidget *widget = gtk_menu_get_attach_widget(data->menu);
+    GdkDisplay* dpy = widget ? gtk_widget_get_display(widget) : gdk_display_get_default();
+    GtkClipboard* clipboard = gtk_clipboard_get_for_display(dpy, GDK_SELECTION_CLIPBOARD);
+    GString *str = g_string_sized_new(128);
+
+    for (fl = fm_file_info_list_peek_head_link(data->file_infos); fl; fl = fl->next)
+    {
+        char *path = fm_path_to_str(fm_file_info_get_path(fl->data));
+        if (str->len > 0)
+            g_string_append_c(str, '\n');
+        g_string_append(str, path);
+        g_free(path);
+    }
+    gtk_clipboard_set_text(clipboard, str->str, str->len);
+    g_string_free(str, TRUE);
+}
+
+static void on_hide(GtkAction* action, FmFileMenu* menu)
+{
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(menu->menu));
+    FmFileInfo *fi = fm_file_info_list_peek_head(menu->file_infos);
+    fm_hide_file(window, fm_file_info_get_path(fi));
+}
+
+static void on_unhide(GtkAction* action, FmFileMenu* menu)
+{
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(menu->menu));
+    FmFileInfo *fi = fm_file_info_list_peek_head(menu->file_infos);
+    fm_unhide_file(window, fm_file_info_get_path(fi));
+}
+
+static void on_add_bookmark(GtkAction* action, FmFileMenu* menu)
+{
+    FmFileInfo* fi = fm_file_info_list_peek_head(menu->file_infos);
+    FmBookmarks* bookmarks;
+    FmPath *path;
+    GList *all, *l;
+    if(fi)
+    {
+        bookmarks = fm_bookmarks_dup();
+        path = fm_file_info_get_path(fi);
+        all = fm_bookmarks_get_all(bookmarks);
+        for (l = all; l; l = l->next)
+            if (path == ((FmBookmarkItem*)l->data)->path)
+                break;
+        g_list_free_full(all, (GDestroyNotify)fm_bookmark_item_unref);
+        if (l == NULL)
+            fm_bookmarks_append(bookmarks, path, fm_file_info_get_disp_name(fi));
+        else
+            fm_show_error(GTK_WINDOW(gtk_menu_get_attach_widget(menu->menu)), NULL,
+                          _("Your bookmarks already have a bookmark for this folder."));
+        g_object_unref(bookmarks);
+    }
 }
 
 static void on_rename(GtkAction* action, gpointer user_data)
 {
     FmFileMenu* data = (FmFileMenu*)user_data;
     FmFileInfo* fi = fm_file_info_list_peek_head(data->file_infos);
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(data->menu));
     if(fi)
-        fm_rename_file(data->parent, fm_file_info_get_path(fi));
-    /* FIXME: is it ok to only rename the first selected file here? */
+        fm_rename_file(window, fm_file_info_get_path(fi));
 }
 
 static void on_compress(GtkAction* action, gpointer user_data)
@@ -831,7 +853,8 @@ static void on_extract_to(GtkAction* action, gpointer user_data)
 static void on_prop(GtkAction* action, gpointer user_data)
 {
     FmFileMenu* data = (FmFileMenu*)user_data;
-    fm_show_file_properties(data->parent, data->file_infos);
+    GtkWindow *window = GTK_WINDOW(gtk_menu_get_attach_widget(data->menu));
+    fm_show_file_properties(window, data->file_infos);
 }
 
 /**
@@ -865,3 +888,45 @@ void fm_file_menu_set_folder_func(FmFileMenu* menu, FmLaunchFolderFunc func, gpo
     menu->folder_func = func;
     menu->folder_func_data = user_data;
 }
+
+/* modules support */
+FM_MODULE_DEFINE_TYPE(gtk_menu_mime, FmFileMenuMimeAddonInit, 1)
+
+static gboolean fm_module_callback_gtk_menu_mime(const char *name, gpointer init, int ver)
+{
+    FmFileMenuMimeExt *ext = g_slice_new(FmFileMenuMimeExt);
+    FmFileMenuMimeAddonInit *cb = init;
+
+    /* not checking version, it's only 1 for now */
+    ext->mime_type = fm_mime_type_from_name(name);
+    ext->cb = *cb;
+    if (cb->init != NULL)
+        cb->init();
+    extensions = g_list_append(extensions, ext);
+    return TRUE;
+}
+
+void _fm_file_menu_init(void)
+{
+    fm_module_register_gtk_menu_mime();
+}
+
+void _fm_file_menu_finalize(void)
+{
+    GList *list, *l;
+    FmFileMenuMimeExt *ext;
+
+    list = extensions;
+    extensions = NULL;
+    for (l = list; l; l = l->next)
+    {
+        ext = l->data;
+        if (ext->cb.finalize != NULL)
+            ext->cb.finalize();
+        if (ext->mime_type != NULL)
+            fm_mime_type_unref(ext->mime_type);
+        g_slice_free(FmFileMenuMimeExt, ext);
+    }
+    g_list_free(list);
+    fm_module_unregister_type("gtk_menu_mime");
+}
