@@ -2,7 +2,7 @@
  *      fm-folder-model.c
  *
  *      Copyright 2009 - 2012 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
- *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2012-2015 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  * @short_description: A model for folder view window.
  * @title: FmFolderModel
  *
- * @include: libfm/fm-folder-model.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmFolderModel is used by widgets such as #FmFolderView to arrange
  * items of folder.
@@ -84,6 +84,8 @@ struct _FmFolderItem
     gboolean is_thumbnail : 1;
     gboolean thumbnail_loading : 1;
     gboolean thumbnail_failed : 1;
+    gboolean is_extra : 1;
+    FmFolderModelExtraFilePos pos : 3;
 };
 
 typedef struct _FmFolderModelFilterItem
@@ -178,6 +180,9 @@ typedef struct
     const char *name;
     const char *title;
     gboolean sortable;
+    gint default_width;
+    void (*get_value)(FmFileInfo *fi, GValue *value);
+    gint (*compare)(FmFileInfo *fi1, FmFileInfo *fi2);
 } FmFolderModelInfo;
 
 static FmFolderModelInfo column_infos_raw[] = {
@@ -189,13 +194,15 @@ static FmFolderModelInfo column_infos_raw[] = {
     { FM_FOLDER_MODEL_COL_OWNER, 0, "owner", N_("Owner"), FALSE },
     { FM_FOLDER_MODEL_COL_MTIME, 0, "mtime", N_("Modified"), TRUE },
     { FM_FOLDER_MODEL_COL_DIRNAME, 0, "dirname", N_("Location"), TRUE },
+    { FM_FOLDER_MODEL_COL_EXT, 0, "ext", N_("Extension"), TRUE },
     /* columns used internally */
     { FM_FOLDER_MODEL_COL_INFO, 0, "info", NULL, TRUE },
     { FM_FOLDER_MODEL_COL_ICON, 0, "icon", NULL, FALSE },
     { FM_FOLDER_MODEL_COL_GICON, 0, "gicon", NULL, FALSE }
 };
 
-static FmFolderModelInfo* column_infos[FM_FOLDER_MODEL_N_COLS];
+static guint column_infos_n = 0;
+static FmFolderModelInfo** column_infos = NULL;
 
 enum {
     ROW_DELETING,
@@ -275,8 +282,6 @@ static void fm_folder_model_class_init(FmFolderModelClass *klass)
 
 static void fm_folder_model_tree_model_init(GtkTreeModelIface *iface)
 {
-    guint i;
-
     iface->get_flags = fm_folder_model_get_flags;
     iface->get_n_columns = fm_folder_model_get_n_columns;
     iface->get_column_type = fm_folder_model_get_column_type;
@@ -289,35 +294,6 @@ static void fm_folder_model_tree_model_init(GtkTreeModelIface *iface)
     iface->iter_n_children = fm_folder_model_iter_n_children;
     iface->iter_nth_child = fm_folder_model_iter_nth_child;
     iface->iter_parent = fm_folder_model_iter_parent;
-
-    /* prepare column_infos table */
-    memset(column_infos, 0, sizeof(column_infos));
-    for(i = 0; i < G_N_ELEMENTS(column_infos_raw); i++)
-    {
-        FmFolderModelCol id = column_infos_raw[i].id;
-        column_infos[id] = &column_infos_raw[i];
-    }
-
-     /* GType value is actually generated at runtime by
-      * calling _get_type() functions for every type.
-      * So they should not be filled at compile-time.
-      * Though G_TYPE_STRING and other fundimental type ids
-      * are known at compile-time, this behavior is not 
-      * guaranteed to be true in newer glib. */
-
-    /* visible columns in the view */
-    column_infos[FM_FOLDER_MODEL_COL_NAME]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_DESC]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_SIZE]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_PERM]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_OWNER]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_MTIME]->type= G_TYPE_STRING;
-    column_infos[FM_FOLDER_MODEL_COL_DIRNAME]->type= G_TYPE_STRING;
-
-    /* columns used internally */
-    column_infos[FM_FOLDER_MODEL_COL_INFO]->type= G_TYPE_POINTER;
-    column_infos[FM_FOLDER_MODEL_COL_ICON]->type= GDK_TYPE_PIXBUF;
-    column_infos[FM_FOLDER_MODEL_COL_GICON]->type= G_TYPE_ICON;
 }
 
 static void fm_folder_model_tree_sortable_init(GtkTreeSortableIface *iface)
@@ -330,14 +306,27 @@ static void fm_folder_model_tree_sortable_init(GtkTreeSortableIface *iface)
     iface->has_default_sort_func = ( gboolean (*)(GtkTreeSortable *) )gtk_false;
 }
 
+static gboolean _drag_source_false(GtkTreeDragSource *drag_source, GtkTreePath *path)
+{ return FALSE; }
+static gboolean _drag_source_false2(GtkTreeDragSource *drag_source, GtkTreePath *path, GtkSelectionData *selection_data)
+{ return FALSE; }
+
 static void fm_folder_model_drag_source_init(GtkTreeDragSourceIface *iface)
 {
-    /* FIXME: Unused. Will this cause any problem? */
+    /* we don't support drag items even for extras - widgets will handle them */
+    iface->row_draggable = _drag_source_false;
+    iface->drag_data_get = _drag_source_false2;
+    iface->drag_data_delete = _drag_source_false;
 }
 
+static gboolean _drag_dest_false(GtkTreeDragDest *drag_dest, GtkTreePath *dest, GtkSelectionData *selection_data)
+{ return FALSE; }
+
 static void fm_folder_model_drag_dest_init(GtkTreeDragDestIface *iface)
 {
-    /* FIXME: Unused. Will this cause any problem? */
+    /* we don't support drag items even for extras - widgets will handle them */
+    iface->drag_data_received = _drag_dest_false;
+    iface->row_drop_possible = _drag_dest_false;
 }
 
 static void fm_folder_model_filter_item_free(FmFolderModelFilterItem* item)
@@ -350,6 +339,16 @@ static void fm_folder_model_dispose(GObject *object)
     FmFolderModel* model = FM_FOLDER_MODEL(object);
     if(model->folder)
         fm_folder_model_set_folder(model, NULL);
+    if(model->items)
+    {
+        g_sequence_free(model->items);
+        model->items = NULL;
+    }
+    if(model->hidden)
+    {
+        g_sequence_free(model->hidden);
+        model->hidden = NULL;
+    }
 
     if(model->theme_change_handler)
     {
@@ -402,6 +401,7 @@ FmFolderModel *fm_folder_model_new(FmFolder* dir, gboolean show_hidden)
     model->hidden = NULL;
     model->show_hidden = show_hidden;
     fm_folder_model_set_folder(model, dir);
+    CHECK_MODULES(); /* prepare columns for application */
     return model;
 }
 
@@ -497,14 +497,49 @@ FmPath* fm_folder_model_get_folder_path(FmFolderModel* model)
  * Changes folder which model handles. This call allows reusing the model
  * for different folder, in case, e.g. directory was changed.
  *
+ * Items added to @model with fm_folder_model_extra_file_add() are not
+ * affected by this API.
+ *
  * Since: 0.1.0
  */
 void fm_folder_model_set_folder(FmFolderModel* model, FmFolder* dir)
 {
+    GSequence *new_items, *new_hidden;
+    GSequenceIter *item_it, *next_item_it;
+    FmFolderItem *item;
+
     if(model->folder == dir)
         return;
     /* g_debug("fm_folder_model_set_folder(%p, %p)", model, dir); */
 
+    new_items = g_sequence_new(fm_folder_item_free);
+    new_hidden = g_sequence_new(fm_folder_item_free);
+    /* keep extra items from removing */
+    if (model->items)
+    {
+        item_it = g_sequence_get_begin_iter(model->items);
+        while(!g_sequence_iter_is_end(item_it))
+        {
+            next_item_it = g_sequence_iter_next(item_it); /* next item */
+            item = (FmFolderItem*)g_sequence_get(item_it);
+            if(item->is_extra)
+                g_sequence_move(item_it, g_sequence_get_end_iter(new_items));
+            item_it = next_item_it;
+        }
+    }
+    if (model->hidden)
+    {
+        item_it = g_sequence_get_begin_iter(model->hidden);
+        while(!g_sequence_iter_is_end(item_it))
+        {
+            next_item_it = g_sequence_iter_next(item_it); /* next item */
+            item = (FmFolderItem*)g_sequence_get(item_it);
+            if(item->is_extra)
+                g_sequence_move(item_it, g_sequence_get_begin_iter(new_hidden));
+            item_it = next_item_it;
+        }
+    }
+
     /* free the old folder */
     if(model->folder)
     {
@@ -539,10 +574,18 @@ void fm_folder_model_set_folder(FmFolderModel* model, FmFolder* dir)
         g_object_unref(model->folder);
         model->folder = NULL;
     }
+    model->items = new_items;
+    model->hidden = new_hidden;
+    /* recreate the hash for items that are in sequence still */
+    item_it = g_sequence_get_begin_iter(model->items);
+    while(!g_sequence_iter_is_end(item_it))
+    {
+        item = (FmFolderItem*)g_sequence_get(item_it);
+        g_hash_table_insert(model->items_hash, item->inf, item_it);
+        item_it = g_sequence_iter_next(item_it);
+    }
     if( !dir )
         return;
-    model->items = g_sequence_new(fm_folder_item_free);
-    model->hidden = g_sequence_new(fm_folder_item_free);
     model->folder = FM_FOLDER(g_object_ref(dir));
 
     g_signal_connect(model->folder, "files-added",
@@ -576,15 +619,17 @@ static GtkTreeModelFlags fm_folder_model_get_flags(GtkTreeModel *tree_model)
 
 static gint fm_folder_model_get_n_columns(GtkTreeModel *tree_model)
 {
-    return FM_FOLDER_MODEL_N_COLS;
+    return (gint)column_infos_n;
 }
 
 static GType fm_folder_model_get_column_type(GtkTreeModel *tree_model,
                                              gint index)
 {
     g_return_val_if_fail(FM_IS_FOLDER_MODEL(tree_model), G_TYPE_INVALID);
-    g_return_val_if_fail(FM_FOLDER_MODEL_COL_IS_VALID(index), G_TYPE_INVALID);
-    g_return_val_if_fail(column_infos[index] != NULL, G_TYPE_INVALID);
+    if((guint)index >= column_infos_n)
+        return G_TYPE_INVALID;
+    if(column_infos[index] == NULL)
+        return G_TYPE_INVALID;
     return column_infos[index]->type;
 }
 
@@ -650,7 +695,7 @@ static void fm_folder_model_get_value(GtkTreeModel *tree_model,
     FmFolderModel* model = FM_FOLDER_MODEL(tree_model);
 
     g_return_if_fail(iter != NULL);
-    g_return_if_fail(FM_FOLDER_MODEL_COL_IS_VALID(column) && column_infos[column] != NULL);
+    g_return_if_fail((guint)column < column_infos_n && column_infos[column] != NULL);
 
     g_value_init(value, column_infos[column]->type);
 
@@ -660,13 +705,19 @@ static void fm_folder_model_get_value(GtkTreeModel *tree_model,
     FmFolderItem* item = (FmFolderItem*)g_sequence_get(item_it);
     FmFileInfo* info = item->inf;
     FmIcon* icon;
+    mode_t mode;
+    char buf[12];
 
-    switch( column )
+    if(column >= FM_FOLDER_MODEL_N_COLS) /* extension */
+    {
+        column_infos[column]->get_value(info, value);
+    }
+    else switch( (FmFolderModelCol)column )
     {
     case FM_FOLDER_MODEL_COL_GICON:
         icon = fm_file_info_get_icon(info);
         if(G_LIKELY(icon))
-            g_value_set_object(value, icon->gicon);
+            g_value_set_object(value, icon);
         break;
     case FM_FOLDER_MODEL_COL_ICON:
     {
@@ -675,13 +726,22 @@ static void fm_folder_model_get_value(GtkTreeModel *tree_model,
             icon = fm_file_info_get_icon(info);
             if(!icon)
                 return;
-            item->icon = fm_pixbuf_from_icon(icon, model->icon_size);
+            /* FIXME: use "emblem-symbolic-link" if file is some kind of link */
+            /* special handle for desktop entries that have invalid icon */
+            if(fm_file_info_is_dir(info))
+                item->icon = fm_pixbuf_from_icon_with_fallback(icon,
+                                                model->icon_size, "folder");
+            else if(fm_file_info_is_desktop_entry(info))
+                item->icon = fm_pixbuf_from_icon_with_fallback(icon,
+                                                model->icon_size, "application-x-executable");
+            else
+                item->icon = fm_pixbuf_from_icon(icon, model->icon_size);
         }
         g_value_set_object(value, item->icon);
 
         /* if we want to show a thumbnail */
         /* if we're on local filesystem or thumbnailing for remote files is allowed */
-        if(fm_config->show_thumbnail && (fm_path_is_local(fm_file_info_get_path(info)) || !fm_config->thumbnail_local))
+        if(fm_config->show_thumbnail && (fm_path_is_native_or_trash(fm_file_info_get_path(info)) || !fm_config->thumbnail_local))
         {
             if(!item->is_thumbnail && !item->thumbnail_failed && !item->thumbnail_loading)
             {
@@ -709,10 +769,42 @@ static void fm_folder_model_get_value(GtkTreeModel *tree_model,
         g_value_set_string( value, fm_file_info_get_desc(info) );
         break;
     case FM_FOLDER_MODEL_COL_PERM:
-//        g_value_set_string( value, fm_file_info_get_disp_perm(info) );
+        mode = fm_file_info_get_mode(info);
+        strcpy(buf, "---------");
+        if ((mode & S_IRUSR) != 0)
+            buf[0] = 'r';
+        if ((mode & S_IWUSR) != 0)
+            buf[1] = 'w';
+        if ((mode & (S_IXUSR | S_ISUID)) == (S_IXUSR | S_ISUID))
+            buf[2] = 's';
+        else if ((mode & S_IXUSR) != 0)
+            buf[2] = 'x';
+        else if ((mode & S_ISUID) != 0)
+            buf[2] = 'S';
+        if ((mode & S_IRGRP) != 0)
+            buf[3] = 'r';
+        if ((mode & S_IWGRP) != 0)
+            buf[4] = 'w';
+        if ((mode & (S_IXGRP | S_ISGID)) == (S_IXGRP | S_ISGID))
+            buf[5] = 's';
+        else if ((mode & S_IXGRP) != 0)
+            buf[5] = 'x';
+        else if ((mode & S_ISGID) != 0)
+            buf[5] = 'S';
+        if ((mode & S_IROTH) != 0)
+            buf[6] = 'r';
+        if ((mode & S_IWOTH) != 0)
+            buf[7] = 'w';
+        if ((mode & (S_IXOTH | S_ISVTX)) == (S_IXOTH | S_ISVTX))
+            buf[8] = 't';
+        else if ((mode & S_IXOTH) != 0)
+            buf[8] = 'x';
+        else if ((mode & S_ISVTX) != 0)
+            buf[8] = 'T';
+        g_value_set_string(value, buf);
         break;
     case FM_FOLDER_MODEL_COL_OWNER:
-//        g_value_set_string( value, fm_file_info_get_disp_owner(info) );
+        g_value_set_string( value, fm_file_info_get_disp_owner(info) );
         break;
     case FM_FOLDER_MODEL_COL_MTIME:
         g_value_set_string( value, fm_file_info_get_disp_mtime(info) );
@@ -733,6 +825,27 @@ static void fm_folder_model_get_value(GtkTreeModel *tree_model,
             }
             break;
         }
+    case FM_FOLDER_MODEL_COL_EXT:
+        {
+            const char *name = fm_file_info_get_disp_name(info);
+            const char *str;
+            if (fm_file_info_is_dir(info))
+            {
+                /* don't look for extension in directories*/
+                str = NULL;
+            }
+            else
+            {
+                str = strrchr(name, '.');
+                if (str == name)
+                    str = NULL;
+                else if (str)
+                    str++;
+            }
+            g_value_set_string(value, str);
+        }
+        break;
+    case FM_FOLDER_MODEL_N_COLS: ; /* unused here */
     }
 }
 
@@ -800,7 +913,7 @@ static gint fm_folder_model_iter_n_children(GtkTreeModel *tree_model,
 {
     FmFolderModel* model;
     g_return_val_if_fail(FM_IS_FOLDER_MODEL(tree_model), -1);
-    g_return_val_if_fail(iter == NULL || iter->user_data != NULL, FALSE);
+    g_return_val_if_fail(iter == NULL || iter->user_data != NULL, -1);
     model = (FmFolderModel*)tree_model;
     /* special case: if iter == NULL, return number of top-level rows */
     if( !iter )
@@ -898,18 +1011,56 @@ static gint fm_folder_model_compare(gconstpointer item1,
     FmFileInfo* file2 = ((FmFolderItem*)item2)->inf;
     const char* key1;
     const char* key2;
+    const char *name;
     goffset diff;
     int ret = 0;
 
     /* put folders before files */
-//    if(model->sort_mode & FM_SORT_FOLDER_FIRST)
+    if(!(model->sort_mode & FM_SORT_NO_FOLDER_FIRST))
     {
         ret = fm_file_info_is_dir(file2) - fm_file_info_is_dir(file1);
         if( ret )
             return ret;
     }
 
-    switch( ((FmFolderModel*)model)->sort_col )
+    /* do with extra items */
+    if (G_UNLIKELY(((FmFolderItem*)item1)->is_extra))
+        switch (((FmFolderItem*)item1)->pos)
+        {
+        case FM_FOLDER_MODEL_ITEMPOS_SORTED:
+            break; /* file1 is in sort sequence */
+        case FM_FOLDER_MODEL_ITEMPOS_PRE:
+            if (((FmFolderItem*)item2)->is_extra &&
+                ((FmFolderItem*)item2)->pos == FM_FOLDER_MODEL_ITEMPOS_PRE)
+                goto _main_sort; /* both are PRE */
+            return -1; /* file1 should be before file2 */
+        case FM_FOLDER_MODEL_ITEMPOS_POST:
+            if (((FmFolderItem*)item2)->is_extra &&
+                ((FmFolderItem*)item2)->pos == FM_FOLDER_MODEL_ITEMPOS_POST)
+                goto _main_sort; /* both are POST */
+            return 1; /* file1 should be after file2 */
+        }
+    if (G_UNLIKELY(((FmFolderItem*)item2)->is_extra))
+        switch (((FmFolderItem*)item2)->pos)
+        {
+        case FM_FOLDER_MODEL_ITEMPOS_SORTED:
+            break; /* file2 is in sort sequence */
+        case FM_FOLDER_MODEL_ITEMPOS_PRE:
+            return 1; /* case if both are PRE was handled above */
+        case FM_FOLDER_MODEL_ITEMPOS_POST:
+            return -1; /* case if both are POST was handled above */
+        }
+
+_main_sort:
+    if(model->sort_col >= FM_FOLDER_MODEL_N_COLS &&
+       model->sort_col < column_infos_n &&
+       column_infos[model->sort_col]->compare)
+    {
+        ret = column_infos[model->sort_col]->compare(file1, file2);
+        if(ret == 0)
+            goto _sort_by_name;
+    }
+    else switch( model->sort_col )
     {
     case FM_FOLDER_MODEL_COL_SIZE:
         /* to support files more than 2Gb */
@@ -943,12 +1094,28 @@ static gint fm_folder_model_compare(gconstpointer item1,
             ret = fm_path_compare(dirpath1, dirpath2);
         }
         break;
+    case FM_FOLDER_MODEL_COL_EXT:
+        name = fm_file_info_get_disp_name(file1);
+        key1 = strrchr(name, '.');
+        if (key1 == name)
+            key1 = NULL;
+        name = fm_file_info_get_disp_name(file2);
+        key2 = strrchr(name, '.');
+        if (key2 == name)
+            key2 = NULL;
+        ret = g_strcmp0(key1, key2);
+        if (ret == 0)
+            goto _sort_by_name;
+        break;
     default:
 _sort_by_name:
         if(model->sort_mode & FM_SORT_CASE_SENSITIVE)
         {
-            key1 = fm_file_info_get_collate_key_nocasefold(file1);
-            key2 = fm_file_info_get_collate_key_nocasefold(file2);
+//            key1 = fm_file_info_get_collate_key_nocasefold(file1);
+//            key2 = fm_file_info_get_collate_key_nocasefold(file2);
+            /* unfortunately g_uft8_collate_key() ignores case in some locales */
+            key1 = fm_file_info_get_disp_name(file1);
+            key2 = fm_file_info_get_disp_name(file2);
         }
         else
         {
@@ -1007,9 +1174,8 @@ static void fm_folder_model_do_sort(FmFolderModel* model)
     g_free(new_order);
 }
 
-static void _fm_folder_model_insert_item(FmFolder* dir,
-                                         FmFolderItem* new_item,
-                                         FmFolderModel* model)
+static void _fm_folder_model_insert_item(FmFolderModel* model,
+                                         FmFolderItem* new_item)
 {
     GtkTreeIter it;
     GtkTreePath* path;
@@ -1037,7 +1203,49 @@ static void _fm_folder_model_insert_item(FmFolder* dir,
 void fm_folder_model_file_created(FmFolderModel* model, FmFileInfo* file)
 {
     FmFolderItem* new_item = fm_folder_item_new(file);
-    _fm_folder_model_insert_item(model->folder, new_item, model);
+    _fm_folder_model_insert_item(model, new_item);
+}
+
+/**
+ * fm_folder_model_extra_file_add
+ * @model: the folder model instance
+ * @file: the file into
+ * @where: position where to put @file in folder model
+ *
+ * Adds @file into @model. Added file will stay at defined position after
+ * any folder change.
+ *
+ * See also: fm_folder_model_set_folder(), fm_folder_model_extra_file_remove().
+ *
+ * Returns: %FALSE if adding was failed.
+ *
+ * Since: 1.2.0
+ */
+gboolean fm_folder_model_extra_file_add(FmFolderModel* model, FmFileInfo* file,
+                                        FmFolderModelExtraFilePos where)
+{
+    FmFolderItem *item;
+
+    /* check visible items first */
+    if (g_hash_table_lookup(model->items_hash, file) != NULL)
+        return FALSE; /* it is already there! */
+    /* check hidden items as well */
+    if (!file_can_show(model, file)) /* if this is a hidden file */
+    {
+        GSequenceIter *seq_it = g_sequence_get_begin_iter(model->hidden);
+        while (!g_sequence_iter_is_end(seq_it))
+        {
+            item = (FmFolderItem*)g_sequence_get(seq_it);
+            if (item->inf == file) /* found! */
+                return FALSE;
+            seq_it = g_sequence_iter_next(seq_it);
+        }
+    }
+    item = fm_folder_item_new(file);
+    item->is_extra = TRUE;
+    item->pos = where;
+    _fm_folder_model_insert_item(model, item);
+    return TRUE;
 }
 
 static inline GSequenceIter* info2iter(FmFolderModel* model, FmFileInfo* file)
@@ -1077,9 +1285,7 @@ void fm_folder_model_file_deleted(FmFolderModel* model, FmFileInfo* file)
         return;
     }
     seq_it = info2iter(model, file);
-    if(!seq_it)
-        /* FIXME: it is an error! */
-        return;
+    g_return_if_fail(seq_it != NULL);
     item = (FmFolderItem*)g_sequence_get(seq_it);
 
     path = gtk_tree_path_new_from_indices(g_sequence_iter_get_position(seq_it), -1);
@@ -1092,6 +1298,62 @@ void fm_folder_model_file_deleted(FmFolderModel* model, FmFileInfo* file)
     g_sequence_remove(seq_it);
 }
 
+/**
+ * fm_folder_model_extra_file_remove
+ * @model: the folder model instance
+ * @file: the file into
+ *
+ * Removes @file from @model if @file was added to the model using
+ * fm_folder_model_extra_file_add().
+ *
+ * Returns: %FALSE if @file cannot be removed.
+ *
+ * Since: 1.2.0
+ */
+gboolean fm_folder_model_extra_file_remove(FmFolderModel* model, FmFileInfo* file)
+{
+    GSequenceIter *seq_it;
+    FmFolderItem *item = NULL;
+    GtkTreePath *path;
+    GtkTreeIter it;
+    gboolean is_hidden = FALSE;
+
+    /* check visible items */
+    seq_it = info2iter(model, file);
+    if (seq_it)
+        item = (FmFolderItem*)g_sequence_get(seq_it);
+    /* check hidden items */
+    else if (!file_can_show(model, file)) /* if this is a hidden file */
+    {
+        seq_it = g_sequence_get_begin_iter(model->hidden);
+        while (!g_sequence_iter_is_end(seq_it))
+        {
+            item = (FmFolderItem*)g_sequence_get(seq_it);
+            if (item->inf == file) /* found! */
+                break;
+            item = NULL; /* reset it again */
+            seq_it = g_sequence_iter_next(seq_it);
+        }
+        is_hidden = TRUE;
+    }
+    if (item == NULL) /* item not found */
+        return FALSE;
+    if (!item->is_extra) /* it wasn't added with fm_folder_model_extra_file_add */
+        return FALSE;
+    if (!is_hidden) /* it is visible, notify everyone we remove it */
+    {
+        path = gtk_tree_path_new_from_indices(g_sequence_iter_get_position(seq_it), -1);
+        it.stamp = model->stamp;
+        it.user_data = seq_it;
+        g_signal_emit(model, signals[ROW_DELETING], 0, path, &it, item->userdata);
+        gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), path);
+        gtk_tree_path_free(path);
+        g_hash_table_remove(model->items_hash, file);
+    }
+    g_sequence_remove(seq_it);
+    return TRUE;
+}
+
 /**
  * fm_folder_model_file_changed
  * @model: a folder model instance
@@ -1108,13 +1370,58 @@ void fm_folder_model_file_changed(FmFolderModel* model, FmFileInfo* file)
     GtkTreeIter it;
     GtkTreePath* path;
 
+    it.stamp = model->stamp;
     if(!file_can_show(model, file))
+    {
+        items_it = info2iter(model, file);
+        if (items_it) /* file was visible and now is hidden */
+        {
+            gint delete_pos = g_sequence_iter_get_position(items_it); /* get row index */
+            it.user_data = items_it; /* setup the tree iterator */
+            g_hash_table_remove(model->items_hash, file);
+            /* move the item from visible list to hidden list */
+            g_sequence_move(items_it, g_sequence_get_begin_iter(model->hidden));
+            /* tell everybody that we removed the item */
+            path = gtk_tree_path_new_from_indices(delete_pos, -1);
+            item = (FmFolderItem*)g_sequence_get(items_it);
+            g_signal_emit(model, signals[ROW_DELETING], 0, path, &it, item->userdata);
+            gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), path);
+            gtk_tree_path_free(path);
+        }
         return;
+    }
 
     items_it = info2iter(model, file);
 
     if(!items_it)
-        return;
+    {
+        /* handle this: file was hidden and now is visible */
+        items_it = g_sequence_get_begin_iter(model->hidden);
+        while(!g_sequence_iter_is_end(items_it)) /* iterate over the hidden list */
+        {
+            item = (FmFolderItem*)g_sequence_get(items_it);
+            if (item->inf == file)
+            {
+                /* find a nice position in visible item list to insert the item */
+                GSequenceIter* insert_item_it = g_sequence_search(model->items, item,
+                                                        fm_folder_model_compare, model);
+                it.user_data  = items_it; /* setup the tree iterator */
+                /* move the item from hidden items to visible items list */
+                g_sequence_move(items_it, insert_item_it);
+                g_hash_table_insert(model->items_hash, file, items_it);
+
+                /* tell the world that we inserted it */
+                path = gtk_tree_path_new_from_indices(g_sequence_iter_get_position(items_it), -1);
+                gtk_tree_model_row_inserted(GTK_TREE_MODEL(model), path, &it);
+                gtk_tree_path_free(path);
+                return;
+            }
+            items_it = g_sequence_iter_next(items_it);
+        }
+        /* item found nowhere, shouldn't we crash? */
+        g_return_if_reached();
+    }
+
     item = (FmFolderItem*)g_sequence_get(items_it);
 
     /* update the icon */
@@ -1122,8 +1429,8 @@ void fm_folder_model_file_changed(FmFolderModel* model, FmFileInfo* file)
     {
         g_object_unref(item->icon);
         item->icon = NULL;
+        item->is_thumbnail = FALSE;
     }
-    it.stamp = model->stamp;
     it.user_data  = items_it;
 
     path = gtk_tree_path_new_from_indices(g_sequence_iter_get_position(items_it), -1);
@@ -1290,23 +1597,24 @@ static void on_thumbnail_loaded(FmThumbnailRequest* req, gpointer user_data)
             GtkTreePath* tp;
             it.stamp = model->stamp;
             it.user_data = seq_it;
+            GDK_THREADS_ENTER();
             tp = fm_folder_model_get_path(GTK_TREE_MODEL(model), &it);
             if(item->icon)
                 g_object_unref(item->icon);
             item->icon = g_object_ref(pix);
             item->is_thumbnail = TRUE;
-            gdk_threads_enter();
             gtk_tree_model_row_changed(GTK_TREE_MODEL(model), tp, &it);
             gtk_tree_path_free(tp);
-            gdk_threads_leave();
+            GDK_THREADS_LEAVE();
         }
         else
         {
             item->thumbnail_failed = TRUE;
         }
         item->thumbnail_loading = FALSE;
+        return;
     }
-    /* FIXME: else should be error? */
+    g_return_if_reached();
 }
 
 /**
@@ -1396,7 +1704,7 @@ static void on_thumbnail_local_changed(FmConfig* cfg, gpointer user_data)
             GList* next = l->next;
             req = (FmThumbnailRequest*)l->data;
             fi = fm_thumbnail_request_get_file_info(req);
-            if(!fm_path_is_local(fm_file_info_get_path(fi)))
+            if(!fm_path_is_native_or_trash(fm_file_info_get_path(fi)))
             {
                 fm_thumbnail_request_cancel(req);
                 model->thumbnail_requests = g_list_delete_link(model->thumbnail_requests, l);
@@ -1414,13 +1722,13 @@ static void on_thumbnail_local_changed(FmConfig* cfg, gpointer user_data)
         if(cfg->thumbnail_local)
         {
             /* add all non-local files to thumbnail requests */
-            if(!fm_path_is_local(path))
+            if(!fm_path_is_native_or_trash(path))
                 reload_thumbnail(model, seq_it, item);
         }
         else
         {
             /* add all non-local files to thumbnail requests */
-            if(!fm_path_is_local(path))
+            if(!fm_path_is_native_or_trash(path))
             {
                 req = fm_thumbnail_request(fi, model->icon_size, on_thumbnail_loaded, model);
                 new_reqs = g_list_append(new_reqs, req);
@@ -1443,31 +1751,6 @@ static void on_thumbnail_max_changed(FmConfig* cfg, gpointer user_data)
     guint thumbnail_max_bytes = fm_config->thumbnail_max << 10;
     goffset size;
 
-#if 0
-    /* disabled due to bug #3557764: pcmanfm-1.0 segfault in fm_file_info_is_image
-       access to FmThumbnailRequest is dangerous in this implementation
-       and thumbnail generator anyway respects cfg->thumbnail_max */
-    if(cfg->thumbnail_max)
-    {
-         /* remove files which are too big from thumbnail requests
-          * FIXME: should this part be put in fm-thumbnail.c instead? */
-        for(l = model->thumbnail_requests; l; )
-        {
-            GList* next = l->next;
-            req = (FmThumbnailRequest*)l->data;
-            fi = fm_thumbnail_request_get_file_info(req);
-            /* call fm_file_info_is_image() here to see if the thumbnail
-             * is generated with built-in GdkPixbuf thumbnailer.
-             * We don't limit file sizes for external thumbnailers */
-            if(fm_file_info_is_image(fi) && fm_file_info_get_size(fi) > (cfg->thumbnail_max << 10))
-            {
-                fm_thumbnail_request_cancel(req);
-                model->thumbnail_requests = g_list_delete_link(model->thumbnail_requests, l);
-            }
-            l = next;
-        }
-    }
-#endif
     seq_it = g_sequence_get_begin_iter(model->items);
     while( !g_sequence_iter_is_end(seq_it) )
     {
@@ -1726,7 +2009,7 @@ void fm_folder_model_set_sort(FmFolderModel* model, FmFolderModelCol col, FmSort
     FmFolderModelCol old_col = model->sort_col;
 
     /* g_debug("fm_folder_model_set_sort: col %x mode %x", col, mode); */
-    if(!FM_FOLDER_MODEL_COL_IS_VALID(col))
+    if((guint)col >= column_infos_n)
         col = old_col;
     if(mode == FM_SORT_DEFAULT)
         mode = model->sort_mode;
@@ -1767,7 +2050,7 @@ gboolean fm_folder_model_get_sort(FmFolderModel* model, FmFolderModelCol *col,
 
 /**
  * fm_folder_model_col_get_title
- * @model: the folder model
+ * @model: (allow-none): the folder model
  * @col_id: column id
  *
  * Retrieves the title of the column specified, or %NULL if the specified
@@ -1780,26 +2063,26 @@ gboolean fm_folder_model_get_sort(FmFolderModel* model, FmFolderModelCol *col,
  */
 const char* fm_folder_model_col_get_title(FmFolderModel* model, FmFolderModelCol col_id)
 {
-    if(G_UNLIKELY(!FM_FOLDER_MODEL_COL_IS_VALID(col_id)
+    if(G_UNLIKELY((guint)col_id >= column_infos_n
        || column_infos[col_id] == NULL)) /* invalid id */
         return NULL;
     return _(column_infos[col_id]->title);
 }
 
 /**
-* fm_folder_model_col_is_sortable
-* @model: model to check
-* @col_id: column id
-*
-* Checks if model can be sorted by @col.
-*
-* Returns: %TRUE if model can be sorted by @col.
-*
-* Since: 1.0.2
-*/
+ * fm_folder_model_col_is_sortable
+ * @model: (allow-none): model to check
+ * @col_id: column id
+ *
+ * Checks if model can be sorted by @col_id.
+ *
+ * Returns: %TRUE if model can be sorted by @col_id.
+ *
+ * Since: 1.0.2
+ */
 gboolean fm_folder_model_col_is_sortable(FmFolderModel* model, FmFolderModelCol col_id)
 {
-    if(G_UNLIKELY(!FM_FOLDER_MODEL_COL_IS_VALID(col_id)
+    if(G_UNLIKELY((guint)col_id >= column_infos_n
        || column_infos[col_id] == NULL)) /* invalid id */
         return FALSE;
     return column_infos[col_id]->sortable;
@@ -1821,7 +2104,7 @@ gboolean fm_folder_model_col_is_sortable(FmFolderModel* model, FmFolderModelCol
  */
 const char* fm_folder_model_col_get_name(FmFolderModelCol col_id)
 {
-    if(G_UNLIKELY(!FM_FOLDER_MODEL_COL_IS_VALID(col_id)
+    if(G_UNLIKELY((guint)col_id >= column_infos_n
        || column_infos[col_id] == NULL)) /* invalid id */
         return NULL;
     return column_infos[col_id]->name;
@@ -1844,7 +2127,7 @@ FmFolderModelCol fm_folder_model_get_col_by_name(const char* str)
     if(G_LIKELY(str != NULL))
     {
         FmFolderModelCol i = 0;
-        for(i = 0; i < G_N_ELEMENTS(column_infos); ++i)
+        for(i = 0; i < column_infos_n; ++i)
         {
             if(column_infos[i] && strcmp(str, column_infos[i]->name) == 0)
                 return i;
@@ -1852,3 +2135,139 @@ FmFolderModelCol fm_folder_model_get_col_by_name(const char* str)
     }
     return (FmFolderModelCol)-1;
 }
+
+/**
+ * fm_folder_model_col_get_default_width
+ * @model: (allow-none): model to check
+ * @col_id: column id
+ *
+ * Retrieves preferred width for @col_id.
+ *
+ * Returns: default width.
+ *
+ * Since: 1.2.0
+ */
+gint fm_folder_model_col_get_default_width(FmFolderModel* model,
+                                           FmFolderModelCol col_id)
+{
+    if(G_UNLIKELY((guint)col_id >= column_infos_n
+       || column_infos[col_id] == NULL)) /* invalid id */
+        return 0;
+    return column_infos[col_id]->default_width;
+}
+
+/**
+ * fm_folder_model_add_custom_column
+ * @name: unique name of column
+ * @init: setup data for column
+ *
+ * Registers custom columns in #FmFolderModel handlers.
+ *
+ * Returns: new column ID or FM_FOLDER_MODEL_COL_DEFAULT in case of failure.
+ *
+ * Since: 1.2.0
+ */
+FmFolderModelCol fm_folder_model_add_custom_column(const char* name,
+                                                   FmFolderModelColumnInit* init)
+{
+    FmFolderModelInfo* info;
+    guint i;
+
+    g_return_val_if_fail(name && init && init->title && init->get_type &&
+                         init->get_value, FM_FOLDER_MODEL_COL_DEFAULT);
+    for(i = 0; i < column_infos_n; i++)
+        if(strcmp(name, column_infos[i]->name) == 0) /* already exists */
+            return FM_FOLDER_MODEL_COL_DEFAULT;
+    column_infos = g_realloc(column_infos, sizeof(FmFolderModelInfo*) * (i+1));
+    info = g_new0(FmFolderModelInfo, 1);
+    column_infos[i] = info;
+    column_infos_n = i+1;
+    info->type = init->get_type();
+    info->name = g_strdup(name);
+    info->title = g_strdup(init->title);
+    info->sortable = (init->compare != NULL);
+    info->default_width = init->default_width;
+    info->get_value = init->get_value;
+    info->compare = init->compare;
+    return i;
+}
+
+/**
+ * fm_folder_model_col_is_valid
+ * @col_id: column id
+ *
+ * Checks if @col_id can be handled by #FmFolderModel.
+ * This API makes things similar to gtk_tree_model_get_n_columns() but it
+ * doesn't operate the model instance.
+ *
+ * Returns: %TRUE if @col_id is valid.
+ *
+ * Since: 1.2.0
+ */
+gboolean fm_folder_model_col_is_valid(FmFolderModelCol col_id)
+{
+    return col_id < column_infos_n;
+}
+
+FM_MODULE_DEFINE_TYPE(gtk_folder_col, FmFolderModelColumnInit, 1)
+
+static gboolean fm_module_callback_gtk_folder_col(const char *name, gpointer init, int ver)
+{
+    if (fm_folder_model_add_custom_column(name, init) == FM_FOLDER_MODEL_COL_DEFAULT)
+        return FALSE;
+    return TRUE;
+}
+
+void _fm_folder_model_init(void)
+{
+    guint i;
+
+    /* prepare column_infos table */
+    column_infos_n = FM_FOLDER_MODEL_N_COLS;
+    column_infos = g_new0(FmFolderModelInfo*, FM_FOLDER_MODEL_N_COLS);
+    for(i = 0; i < G_N_ELEMENTS(column_infos_raw); i++)
+    {
+        FmFolderModelCol id = column_infos_raw[i].id;
+        column_infos[id] = &column_infos_raw[i];
+    }
+
+     /* GType value is actually generated at runtime by
+      * calling _get_type() functions for every type.
+      * So they should not be filled at compile-time.
+      * Though G_TYPE_STRING and other fundimental type ids
+      * are known at compile-time, this behavior is not 
+      * guaranteed to be true in newer glib. */
+
+    /* visible columns in the view */
+    column_infos[FM_FOLDER_MODEL_COL_NAME]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_DESC]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_SIZE]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_PERM]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_OWNER]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_MTIME]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_DIRNAME]->type= G_TYPE_STRING;
+    column_infos[FM_FOLDER_MODEL_COL_EXT]->type= G_TYPE_STRING;
+
+    /* columns used internally */
+    column_infos[FM_FOLDER_MODEL_COL_INFO]->type= G_TYPE_POINTER;
+    column_infos[FM_FOLDER_MODEL_COL_ICON]->type= GDK_TYPE_PIXBUF;
+    column_infos[FM_FOLDER_MODEL_COL_GICON]->type= G_TYPE_ICON;
+
+    fm_module_register_gtk_folder_col();
+}
+
+void _fm_folder_model_finalize(void)
+{
+    FmFolderModelCol i = column_infos_n;
+    fm_module_unregister_type("gtk_folder_col");
+    column_infos_n = 0;
+    /* free all custom columns! */
+    while (i > FM_FOLDER_MODEL_N_COLS)
+    {
+        i--;
+        g_free((char*)column_infos[i]->name);
+        g_free((char*)column_infos[i]->title);
+        g_free(column_infos[i]);
+    }
+    g_free(column_infos);
+}
