@@ -3,6 +3,7 @@
  *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
  *      Copyright 2009 Jürgen Hötzel <juergen@archlinux.org>
+ *      Copyright 2012-2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -25,7 +26,7 @@
  * @short_description: An entry to enter path with completion.
  * @title: FmPathEntry
  *
- * @include: libfm/fm-path-entry.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmPathEntry represents a widget to enter folder path for changing
  * current directory. The path entry supports completion and can be used
@@ -34,7 +35,13 @@
  * (such as \%23) into entry.
  */
 
-#include "gtk-compat.h"
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#include <glib/gi18n-lib.h>
+
+#include "../gtk-compat.h"
 
 #include "fm-path-entry.h"
 /* for completion */
@@ -45,6 +52,17 @@
 #include <string.h>
 #include <gio/gio.h>
 #include <gdk/gdkkeysyms.h>
+
+struct _FmPathEntry
+{
+  GtkEntry parent_instance;
+};
+
+struct _FmPathEntryClass
+{
+  GtkEntryClass parent_class;
+};
+
 /* properties */
 enum
 {
@@ -71,6 +89,7 @@ struct _FmPathEntryPrivate
 
     gboolean folder_loaded : 1;
     gboolean highlight_completion_match : 1;
+    gboolean long_list : 1;
     //gboolean complete_on_load :1;
     GtkEntryCompletion* completion;
     guint id_changed;
@@ -181,6 +200,29 @@ static void fm_path_entry_dispatch_properties_changed(GObject *object,
 }
 #endif
 
+static gboolean _path_entry_is_single_match(FmPathEntry *entry, FmPathEntryPrivate *priv)
+{
+    GtkTreeModel *model = GTK_TREE_MODEL(priv->model);
+    char *model_basename;
+    const char* typed_basename;
+    GtkTreeIter it;
+    gboolean partial = FALSE, match = FALSE;
+
+    typed_basename = gtk_entry_get_text(GTK_ENTRY(entry)) + priv->parent_len;
+    if (gtk_tree_model_get_iter_first(model, &it)) do
+    {
+        gtk_tree_model_get(model, &it, COL_BASENAME, &model_basename, -1);
+        if (!match && strcmp(model_basename, typed_basename) == 0)
+            match = TRUE; /* exact match */
+        else if (g_str_has_prefix(model_basename, typed_basename))
+            partial = TRUE;
+        g_free(model_basename);
+    }
+    while (!partial && gtk_tree_model_iter_next(model, &it));
+
+    return (match && !partial);
+}
+
 static gboolean fm_path_entry_key_press(GtkWidget   *widget, GdkEventKey *event, gpointer user_data)
 {
     FmPathEntry *entry = FM_PATH_ENTRY(widget);
@@ -204,6 +246,12 @@ static gboolean fm_path_entry_key_press(GtkWidget   *widget, GdkEventKey *event,
 #endif
             gtk_entry_completion_insert_prefix(priv->completion);
             gtk_editable_set_position(GTK_EDITABLE(entry), -1);
+            if (_path_entry_is_single_match(entry, priv))
+            {
+                int pos = gtk_editable_get_position(GTK_EDITABLE(entry));
+                gtk_editable_insert_text(GTK_EDITABLE(entry), "/", 1, &pos);
+                gtk_editable_set_position(GTK_EDITABLE(entry), pos);
+            }
             return TRUE;
         }
     }
@@ -212,18 +260,10 @@ static gboolean fm_path_entry_key_press(GtkWidget   *widget, GdkEventKey *event,
 
 static void _set_entry_text_from_path(GtkEntry *entry, FmPathEntryPrivate *priv)
 {
-    char *disp_name, *escaped;
+    char *disp_name;
     FmPath *path = priv->path;
 
-    if(fm_path_is_native(path))
-        disp_name = fm_path_display_name(path, FALSE);
-    else
-    {
-        /* URIs are escaped there */
-        escaped = fm_path_to_str(path);
-        disp_name = g_uri_unescape_string(escaped, NULL);
-        g_free(escaped);
-    }
+    disp_name = fm_path_display_name(path, FALSE);
     /* block our handler for "changed" signal, we'll update it below */
     if(priv->id_changed > 0)
         g_signal_handler_block(entry, priv->id_changed);
@@ -251,7 +291,15 @@ static void fm_path_entry_activate(GtkEntry *entry, gpointer user_data)
     else if(full_path[0] == '~' && full_path[1] == 0)
         priv->path = fm_path_ref(fm_path_get_home());
     else
-        priv->path = fm_path_new_for_str(full_path);
+        priv->path = fm_path_new_for_display_name(full_path);
+
+#if G_ENABLE_DEBUG
+    {
+        char *real_path = fm_path_to_str(priv->path);
+        g_debug("FmPathEntry activated: '%s' => '%s'", full_path, real_path);
+        g_free(real_path);
+    }
+#endif
 
     _set_entry_text_from_path(entry, priv);
 
@@ -351,7 +399,12 @@ static gboolean fm_path_entry_focus_out_event(GtkWidget *widget, GdkEventFocus *
     return GTK_WIDGET_CLASS(fm_path_entry_parent_class)->focus_out_event(widget, event);
 }
 
+#if GLIB_CHECK_VERSION(2, 36, 0)
+static void on_dir_list_finished(GObject *source_object, GAsyncResult *res,
+                                 gpointer user_data)
+#else
 static gboolean on_dir_list_finished(gpointer user_data)
+#endif
 {
     ListSubDirNames* data = (ListSubDirNames*)user_data;
     FmPathEntry* entry = data->entry;
@@ -361,7 +414,11 @@ static gboolean on_dir_list_finished(gpointer user_data)
 
     /* final chance to check cancellable */
     if(g_cancellable_is_cancelled(data->cancellable))
+#if GLIB_CHECK_VERSION(2, 36, 0)
+        return;
+#else
         return TRUE;
+#endif
     /* FIXME: check errors! */
 
     new_model = fm_path_entry_model_new(priv->parent_dir);
@@ -374,6 +431,7 @@ static gboolean on_dir_list_finished(gpointer user_data)
         gtk_list_store_insert_with_values((GtkListStore*)new_model, NULL, -1, COL_BASENAME, name, -1);
     }
     priv->folder_loaded = TRUE;
+    priv->long_list = (g_list_length(data->subdirs) > 40);
 
     gtk_entry_completion_set_model(priv->completion, GTK_TREE_MODEL(new_model));
     if(priv->model)
@@ -395,19 +453,28 @@ static gboolean on_dir_list_finished(gpointer user_data)
     /* trigger completion popup. FIXME: this is a little bit dirty.
      * A even more dirty thing to do is to check if we finished after
      * 300 ms timeout happens. */
+    g_signal_emit_by_name(entry, "changed", 0);
+#if !GLIB_CHECK_VERSION(2, 36, 0)
     return TRUE;
+#endif
 }
 
+#if GLIB_CHECK_VERSION(2, 36, 0)
+static void list_sub_dirs(GTask *task, gpointer source_object, gpointer user_data,
+                          GCancellable *cancellable)
+#else
 static gboolean list_sub_dirs(GIOSchedulerJob *job, GCancellable *cancellable, gpointer user_data)
+#endif
 {
     ListSubDirNames* data = (ListSubDirNames*)user_data;
     GError *err = NULL;
     /* g_debug("new dir listing job!"); */
     GFileEnumerator* enu = g_file_enumerate_children(data->dir,
-                                    G_FILE_ATTRIBUTE_STANDARD_NAME","
+                                    G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME","
+                                    G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME","
                                     G_FILE_ATTRIBUTE_STANDARD_TYPE,
                                     G_FILE_QUERY_INFO_NONE, cancellable,
-                                    &err);
+                                    NULL);
     if(enu)
     {
         while(!g_cancellable_is_cancelled(cancellable))
@@ -418,7 +485,9 @@ static gboolean list_sub_dirs(GIOSchedulerJob *job, GCancellable *cancellable, g
                 GFileType type = g_file_info_get_file_type(inf);
                 if(type == G_FILE_TYPE_DIRECTORY)
                 {
-                    const char* name = g_file_info_get_name(inf);
+                    const char* name = g_file_info_get_edit_name(inf);
+                    if (!name)
+                        name = g_file_info_get_display_name(inf);
                     data->subdirs = g_list_prepend(data->subdirs, g_strdup(name));
                 }
                 g_object_unref(inf);
@@ -426,10 +495,7 @@ static gboolean list_sub_dirs(GIOSchedulerJob *job, GCancellable *cancellable, g
             else
             {
                 if(err) /* error happens */
-                {
-                    g_error_free(err);
-                    err = NULL;
-                }
+                    g_clear_error(&err);
                 else /* EOF */
                     break;
             }
@@ -438,11 +504,15 @@ static gboolean list_sub_dirs(GIOSchedulerJob *job, GCancellable *cancellable, g
     }
 
     if(!g_cancellable_is_cancelled(cancellable))
+#if GLIB_CHECK_VERSION(2, 36, 0)
+        g_task_return_pointer(task, NULL, NULL);
+#else
     {
         /* finished! */
         g_io_scheduler_job_send_to_mainloop(job, on_dir_list_finished, data, NULL);
     }
     return FALSE;
+#endif
 }
 
 static void list_sub_dir_names_free(gpointer user_data)
@@ -460,8 +530,12 @@ static void fm_path_entry_changed(GtkEditable *editable, gpointer user_data)
     FmPathEntry *entry = FM_PATH_ENTRY(editable);
     FmPathEntryPrivate *priv  = FM_PATH_ENTRY_GET_PRIVATE(entry);
     const gchar *path_str, *sep;
+#if GLIB_CHECK_VERSION(2, 36, 0)
+    GTask *task;
+#endif
 
-    g_return_if_fail(priv->model != NULL);
+    if(priv->model == NULL)
+        return;
     /* find parent dir of current path */
     path_str = gtk_entry_get_text( GTK_ENTRY(entry) );
     sep = g_utf8_strrchr(path_str, -1, G_DIR_SEPARATOR);
@@ -490,14 +564,27 @@ static void fm_path_entry_changed(GtkEditable *editable, gpointer user_data)
                 g_free(expand);
             }
             else
-                data->dir = fm_file_new_for_commandline_arg(priv->parent_dir);
+            {
+                FmPath *p = fm_path_new_for_display_name(priv->parent_dir);
+                data->dir = fm_path_to_gfile(p);
+                fm_path_unref(p);
+            }
 
             /* launch a new job to do dir listing */
-            data->cancellable = g_cancellable_new();
-            priv->cancellable = (GCancellable*)g_object_ref(data->cancellable);
+            if (G_LIKELY(priv->cancellable == NULL))
+                priv->cancellable = g_cancellable_new();
+            data->cancellable = (GCancellable*)g_object_ref(priv->cancellable);
+#if GLIB_CHECK_VERSION(2, 36, 0)
+            task = g_task_new(editable, data->cancellable, on_dir_list_finished, data);
+            g_task_set_task_data(task, data, list_sub_dir_names_free);
+            g_task_set_priority(task, G_PRIORITY_LOW);
+            g_task_run_in_thread(task, list_sub_dirs);
+            g_object_unref(task);
+#else
             g_io_scheduler_push_job(list_sub_dirs,
                                     data, list_sub_dir_names_free,
                                     G_PRIORITY_LOW, data->cancellable);
+#endif
         }
         /* calculate the length of remaining part after / */
         priv->typed_basename_len = strlen(sep + 1);
@@ -543,12 +630,68 @@ static void fm_path_entry_get_property(GObject *object,
     }
 }
 
+static void fm_path_entry_paste_and_go(GtkMenuItem *menuitem, GtkEntry *entry)
+{
+    GtkClipboard* clipboard = gtk_clipboard_get_for_display(
+        gtk_widget_get_display(GTK_WIDGET(menuitem)), GDK_SELECTION_CLIPBOARD);
+
+    gchar* full_path = gtk_clipboard_wait_for_text(clipboard);
+
+    if (full_path)
+    {
+        FmPathEntryPrivate *priv  = FM_PATH_ENTRY_GET_PRIVATE(entry);
+
+        if(priv->path)
+            fm_path_unref(priv->path);
+
+        /* special handling for home dir */
+        if(full_path[0] == '~' && full_path[1] == G_DIR_SEPARATOR)
+            priv->path = fm_path_new_relative(fm_path_get_home(), full_path + 2);
+        else if(full_path[0] == '~' && full_path[1] == 0)
+            priv->path = fm_path_ref(fm_path_get_home());
+        else
+            /* FIXME: use fm_path_new_for_display_name ? */
+            priv->path = fm_path_new_for_str(full_path);
+
+        gchar * disp_name = fm_path_display_name(priv->path, FALSE);
+        gtk_entry_set_text(entry, disp_name);
+        g_free(disp_name);
+
+        gtk_editable_set_position(GTK_EDITABLE(entry), -1);
+
+        g_free(full_path);
+
+        g_signal_emit_by_name(entry, "activate", 0);
+    }
+}
+
+static void fm_path_entry_populate_popup(GtkEntry *entry, GtkMenu *menu, gpointer user_data)
+{
+    GtkWidget* menuitem;
+
+    GtkClipboard* clipboard = gtk_clipboard_get_for_display(
+        gtk_widget_get_display(GTK_WIDGET(entry)), GDK_SELECTION_CLIPBOARD);
+
+    menuitem = gtk_menu_item_new_with_mnemonic(_("Pa_ste and Go"));
+    gtk_widget_show(menuitem);
+
+    /* Insert menu item after default Paste menu item */
+    gtk_menu_shell_insert(GTK_MENU_SHELL(menu), menuitem, 3);
+
+    g_signal_connect(menuitem, "activate",
+                     G_CALLBACK(fm_path_entry_paste_and_go), entry);
+
+    if (!gtk_clipboard_wait_is_text_available(clipboard))
+        gtk_widget_set_sensitive(menuitem, FALSE);
+}
+
 static void
 fm_path_entry_init(FmPathEntry *entry)
 {
     FmPathEntryPrivate *priv = FM_PATH_ENTRY_GET_PRIVATE(entry);
     GtkEntryCompletion* completion = gtk_entry_completion_new();
     GtkCellRenderer* render;
+    AtkObject *obj;
 
     priv->model = fm_path_entry_model_new(NULL);
     priv->completion = completion;
@@ -585,6 +728,10 @@ fm_path_entry_init(FmPathEntry *entry)
      * we want to invoke our handlers before the default ones provided by Gtk. */
     g_signal_connect(entry, "key-press-event", G_CALLBACK(fm_path_entry_key_press), NULL);
     g_signal_connect(entry, "activate", G_CALLBACK(fm_path_entry_activate), NULL);
+    g_signal_connect(entry, "populate-popup", G_CALLBACK(fm_path_entry_populate_popup), NULL);
+
+    obj = gtk_widget_get_accessible(GTK_WIDGET(entry));
+    atk_object_set_description(obj, _("Folder location bar"));
 }
 
 static void fm_path_entry_completion_render_func(GtkCellLayout *cell_layout,
@@ -599,7 +746,7 @@ static void fm_path_entry_completion_render_func(GtkCellLayout *cell_layout,
     gtk_tree_model_get(model, iter, COL_BASENAME, &model_file_name, -1);
     model_file_name_len = strlen(model_file_name);
 
-    if( priv->highlight_completion_match )
+    if( priv->highlight_completion_match && (model_file_name_len >= priv->typed_basename_len) )
     {
         int buf_len = model_file_name_len + 14 + 1;
         gchar* markup = g_malloc(buf_len);
@@ -723,10 +870,13 @@ static gboolean fm_path_entry_match_func(GtkEntryCompletion   *completion,
     typed_basename = gtk_entry_get_text(GTK_ENTRY(entry)) + priv->parent_len;
     gtk_tree_model_get(model, iter, COL_BASENAME, &model_basename, -1);
 
-    if(model_basename[0] == '.' && typed_basename[0] != '.')
+    if (G_UNLIKELY(model_basename == NULL))
+        ret = FALSE; /* it is invalid if file name is empty but it's possible! */
+    else if(model_basename[0] == '.' && typed_basename[0] != '.')
         ret = FALSE; /* ignore hidden files when not requested. */
-    else if(typed_basename[0] == '\0' /* "/xxx/" - no names here yet */
-           || (typed_basename[0] == '.' && typed_basename[1] == '\0')) /* "/xxx/." */
+    else if(priv->long_list && /* don't create too long lists */
+            (typed_basename[0] == '\0' /* "/xxx/" - no names here yet */
+             || (typed_basename[0] == '.' && typed_basename[1] == '\0'))) /* "/xxx/." */
         ret = FALSE;
     else
         ret = g_str_has_prefix(model_basename, typed_basename); /* FIXME: should we be case insensitive here? */
