@@ -2,7 +2,8 @@
  *      fm-places-view.c
  *
  *      Copyright 2009 - 2012 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
- *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2012-2015 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2024 Ingo Brückl
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -25,7 +26,7 @@
  * @short_description: A widget for side panel with places list.
  * @title: FmPlacesView
  *
- * @include: libfm/fm-places-view.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmPlacesView displays list of pseudo-folders which contains
  * such items as Home directory, Trash bin, mounted removable drives,
@@ -36,19 +37,34 @@
 #include <config.h>
 #endif
 
+#define FM_DISABLE_SEAL
+
 #include <glib/gi18n-lib.h>
 #include "fm-places-view.h"
 #include "fm-config.h"
+#include "fm-utils.h"
 #include "fm-gtk-utils.h"
 #include "fm-bookmarks.h"
 #include "fm-file-menu.h"
 #include "fm-cell-renderer-pixbuf.h"
 #include "fm-dnd-auto-scroll.h"
 #include "fm-places-model.h"
+#include "fm-gtk-file-launcher.h"
+#include "fm-gtk-marshal.h"
+
+#include <gdk/gdkkeysyms.h>
+#include "gtk-compat.h"
+
+enum
+{
+    PROP_0,
+    PROP_HOME_DIR
+};
 
 enum
 {
     CHDIR,
+    ITEM_POPUP,
     N_SIGNALS
 };
 
@@ -60,9 +76,12 @@ static gboolean on_button_release(GtkWidget* view, GdkEventButton* evt);
 static void on_mount(GtkAction* act, gpointer user_data);
 static void on_umount(GtkAction* act, gpointer user_data);
 static void on_eject(GtkAction* act, gpointer user_data);
+static void on_format(GtkAction* act, gpointer user_data);
 
 static void on_remove_bm(GtkAction* act, gpointer user_data);
 static void on_rename_bm(GtkAction* act, gpointer user_data);
+static void on_move_bm_up(GtkAction* act, gpointer user_data);
+static void on_move_bm_down(GtkAction* act, gpointer user_data);
 static void on_empty_trash(GtkAction* act, gpointer user_data);
 
 static gboolean on_dnd_dest_files_dropped(FmDndDest* dd, int x, int y, GdkDragAction action,
@@ -80,16 +99,28 @@ static FmPlacesModel* model = NULL;
 
 static guint signals[N_SIGNALS];
 
+#define PLACES_MENU_XML \
+"<popup>" \
+  "<placeholder name='ph1'/>" \
+  "<separator/>" \
+  "<placeholder name='ph2'/>" \
+  "<separator/>" \
+  "<placeholder name='ph3'/>" \
+"</popup>"
+
 static const char vol_menu_xml[]=
+PLACES_MENU_XML
 "<popup>"
   "<placeholder name='ph3'>"
   "<menuitem action='Mount'/>"
   "<menuitem action='Unmount'/>"
   "<menuitem action='Eject'/>"
+  "<menuitem action='Format'/>"
   "</placeholder>"
 "</popup>";
 
 static const char mount_menu_xml[]=
+PLACES_MENU_XML
 "<popup>"
   "<placeholder name='ph3'>"
   "<menuitem action='Unmount'/>"
@@ -100,24 +131,31 @@ static GtkActionEntry vol_menu_actions[]=
 {
     {"Mount", NULL, N_("_Mount Volume"), NULL, NULL, G_CALLBACK(on_mount)},
     {"Unmount", NULL, N_("_Unmount Volume"), NULL, NULL, G_CALLBACK(on_umount)},
-    {"Eject", NULL, N_("_Eject Removable Media"), NULL, NULL, G_CALLBACK(on_eject)}
+    {"Eject", NULL, N_("_Eject Removable Media"), NULL, NULL, G_CALLBACK(on_eject)},
+    {"Format", NULL, N_("_Format Volume"), NULL, NULL, G_CALLBACK(on_format)}
 };
 
 static const char bookmark_menu_xml[]=
+PLACES_MENU_XML
 "<popup>"
   "<placeholder name='ph3'>"
   "<menuitem action='RenameBm'/>"
   "<menuitem action='RemoveBm'/>"
+  "<menuitem action='MoveBmUp'/>"
+  "<menuitem action='MoveBmDown'/>"
   "</placeholder>"
 "</popup>";
 
 static GtkActionEntry bm_menu_actions[]=
 {
     {"RenameBm", GTK_STOCK_EDIT, N_("_Rename Bookmark Item"), NULL, NULL, G_CALLBACK(on_rename_bm)},
-    {"RemoveBm", GTK_STOCK_REMOVE, N_("Re_move from Bookmark"), NULL, NULL, G_CALLBACK(on_remove_bm)}
+    {"RemoveBm", GTK_STOCK_REMOVE, N_("Re_move from Bookmarks"), NULL, NULL, G_CALLBACK(on_remove_bm)},
+    {"MoveBmUp", GTK_STOCK_GO_UP, N_("Move Bookmark _Up"), NULL, NULL, G_CALLBACK(on_move_bm_up)},
+    {"MoveBmDown", GTK_STOCK_GO_DOWN, N_("Move Bookmark _Down"), NULL, NULL, G_CALLBACK(on_move_bm_down)}
 };
 
 static const char trash_menu_xml[]=
+PLACES_MENU_XML
 "<popup>"
   "<placeholder name='ph3'>"
   "<menuitem action='EmptyTrash'/>"
@@ -210,7 +248,7 @@ static gboolean on_drag_motion (GtkWidget *dest_widget,
     /* fm_drag_context_has_target_name(drag_context, "GTK_TREE_MODEL_ROW"); */
     GdkAtom target;
     GtkTreeViewDropPosition pos;
-    GtkTreePath* tp;
+    GtkTreePath* tp = NULL;
     gboolean ret = FALSE;
     GdkDragAction action = 0;
 
@@ -250,7 +288,9 @@ static gboolean on_drag_motion (GtkWidget *dest_widget,
             fm_dnd_dest_set_dest_file(view->dnd_dest, NULL);
             /* FmDndDest requires this call */
             fm_dnd_dest_get_default_action(view->dnd_dest, drag_context, target);
-            if( (!tp || fm_places_model_path_is_bookmark(model, tp))
+            if(!gtk_drag_get_source_widget(drag_context))
+                action = 0; /* an other application can't link */
+            else if( (!tp || fm_places_model_path_is_bookmark(model, tp))
                && get_bookmark_drag_dest(view, &tp, &pos)) /* tp is after separator */
                 action = GDK_ACTION_LINK;
             else
@@ -442,6 +482,7 @@ static void fm_places_view_init(FmPlacesView *self)
 {
     GtkTreeViewColumn* col;
     GtkCellRenderer* renderer;
+    AtkObject *obj;
     guint handler;
 
     if(G_UNLIKELY(!model))
@@ -491,6 +532,8 @@ static void fm_places_view_init(FmPlacesView *self)
     fm_dnd_dest_add_targets(GTK_WIDGET(self), dnd_targets, G_N_ELEMENTS(dnd_targets));
 
     g_signal_connect(self->dnd_dest, "files-dropped", G_CALLBACK(on_dnd_dest_files_dropped), self);
+    obj = gtk_widget_get_accessible(GTK_WIDGET(self));
+    atk_object_set_description(obj, _("Shows list of common places, devices, and bookmarks in sidebar"));
 }
 
 /*----------------------------------------------------------------------
@@ -524,7 +567,11 @@ static void activate_row(FmPlacesView* view, guint button, GtkTreePath* tree_pat
         {
         case FM_PLACES_ITEM_PATH:
         case FM_PLACES_ITEM_MOUNT:
-            path = fm_path_ref(fm_places_item_get_path(item));
+            path = fm_places_item_get_path(item);
+            if (path == fm_path_get_home() && view->home_dir)
+                path = fm_path_new_for_str(view->home_dir);
+            else
+                fm_path_ref(path);
             break;
         case FM_PLACES_ITEM_VOLUME:
         {
@@ -696,11 +743,16 @@ static void place_item_menu_unref(gpointer ui, GObject *menu)
     g_object_unref(ui);
 }
 
-static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget* widget)
+static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget *widget)
 {
     GtkWidget* menu = NULL;
     GtkUIManager* ui = gtk_ui_manager_new();
     GtkActionGroup* act_grp = gtk_action_group_new("Popup");
+    int sep_pos;
+    GtkTreeIter src_it;
+    GtkTreePath *tp, *sep_tp;
+    GtkAction* act;
+
     gtk_action_group_set_translation_domain(act_grp, GETTEXT_PACKAGE);
 
     /* FIXME: merge with FmFileMenu when possible */
@@ -710,6 +762,28 @@ static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget* widget)
         {
             gtk_action_group_add_actions(act_grp, bm_menu_actions, G_N_ELEMENTS(bm_menu_actions), item);
             gtk_ui_manager_add_ui_from_string(ui, bookmark_menu_xml, -1, NULL);
+            /* check and disable MoveBmUp and MoveBmDown */
+            if(fm_places_model_get_iter_by_fm_path(model, &src_it,
+                                                   fm_places_item_get_path(item)))
+            {
+                /* get index of the separator */
+                sep_tp = fm_places_model_get_separator_path(model);
+                sep_pos = gtk_tree_path_get_indices(sep_tp)[0];
+                tp = gtk_tree_model_get_path(GTK_TREE_MODEL(model), &src_it);
+                if(!gtk_tree_path_prev(tp) ||
+                   gtk_tree_path_get_indices(tp)[0] - sep_pos - 1 < 0)
+                {
+                    act = gtk_action_group_get_action(act_grp, "MoveBmUp");
+                    gtk_action_set_sensitive(act, FALSE);
+                }
+                if(!gtk_tree_model_iter_next(GTK_TREE_MODEL(model), &src_it))
+                {
+                    act = gtk_action_group_get_action(act_grp, "MoveBmDown");
+                    gtk_action_set_sensitive(act, FALSE);
+                }
+                gtk_tree_path_free(sep_tp);
+                gtk_tree_path_free(tp);
+            }
         }
         else if(fm_path_is_trash_root(fm_places_item_get_path(item)))
         {
@@ -719,22 +793,42 @@ static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget* widget)
     }
     else if(fm_places_item_get_type(item) == FM_PLACES_ITEM_VOLUME)
     {
-        GtkAction* act;
+        GVolume *vol = fm_places_item_get_volume(item);
         GMount* mnt;
+        char *unix_path = NULL;
         gtk_action_group_add_actions(act_grp, vol_menu_actions, G_N_ELEMENTS(vol_menu_actions), item);
         gtk_ui_manager_add_ui_from_string(ui, vol_menu_xml, -1, NULL);
 
-        mnt = g_volume_get_mount(fm_places_item_get_volume(item));
+        mnt = g_volume_get_mount(vol);
         if(mnt) /* mounted */
         {
             g_object_unref(mnt);
             act = gtk_action_group_get_action(act_grp, "Mount");
             gtk_action_set_visible(act, FALSE);
+            act = gtk_action_group_get_action(act_grp, "Unmount");
+            gtk_action_set_sensitive(act, g_mount_can_unmount(mnt));
         }
         else /* not mounted */
         {
+            if (fm_config->format_cmd && fm_config->format_cmd[0])
+                unix_path = g_volume_get_identifier(vol,
+                                                    G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
+            if (unix_path && unix_path[0] != '/') /* we can format only local */
+            {
+                g_free(unix_path);
+                unix_path = NULL;
+            }
+            g_free(unix_path); /* use it to mark only */
             act = gtk_action_group_get_action(act_grp, "Unmount");
             gtk_action_set_visible(act, FALSE);
+            act = gtk_action_group_get_action(act_grp, "Mount");
+            gtk_action_set_sensitive(act, g_volume_can_mount(vol));
+        }
+        if (unix_path == NULL)
+        {
+            act = gtk_action_group_get_action(act_grp, "Format");
+            if (act)
+                 gtk_action_set_visible(act, FALSE);
         }
 
         if(!g_volume_can_eject(fm_places_item_get_volume(item)))
@@ -755,17 +849,27 @@ static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget* widget)
         {
             act = gtk_action_group_get_action(act_grp, "Mount");
             gtk_action_set_sensitive(act, FALSE);
+            act = gtk_action_group_get_action(act_grp, "Unmount");
+            gtk_action_set_sensitive(act, g_mount_can_unmount(mnt));
         }
         else /* not mounted */
         {
             act = gtk_action_group_get_action(act_grp, "Unmount");
             gtk_action_set_sensitive(act, FALSE);
         }
+        act = gtk_action_group_get_action(act_grp, "Format");
+        if (act)
+             gtk_action_set_visible(act, FALSE);
+        act = gtk_action_group_get_action(act_grp, "Eject");
+        gtk_action_set_visible(act, FALSE);
     }
     else
         goto _out;
     gtk_ui_manager_insert_action_group(ui, act_grp, 0);
 
+    /* send the signal so popup can be altered by application */
+    g_signal_emit(widget, signals[ITEM_POPUP], 0, ui, act_grp, fm_places_item_get_info(item));
+
     menu = gtk_ui_manager_get_widget(ui, "/popup");
     if(menu)
     {
@@ -776,6 +880,7 @@ static GtkWidget* place_item_get_menu(FmPlacesItem* item, GtkWidget* widget)
            Unmount therefore we should detach menu to avoid crash
            note that we should remove this ref when we destroy menu */
         g_object_weak_ref(G_OBJECT(widget), (GWeakNotify)gtk_menu_detach, menu);
+        gtk_ui_manager_ensure_update(ui);
     }
 
 _out:
@@ -784,11 +889,95 @@ _out:
     return menu;
 }
 
+static void popup_position_func(GtkMenu *menu, gint *x, gint *y,
+                                gboolean *push_in, gpointer user_data)
+{
+    GtkWidget *widget = gtk_menu_get_attach_widget(menu);
+    GtkTreeView *view = GTK_TREE_VIEW(widget);
+    GtkTreePath *path;
+    GdkScreen *screen;
+    gint index = GPOINTER_TO_INT(user_data);
+    GdkRectangle cell;
+    GtkAllocation a, ma;
+    gint x2, y2, mon;
+    gboolean rtl = (gtk_widget_get_direction(widget) == GTK_TEXT_DIR_RTL);
+
+    /* realize menu so we get actual size of it */
+    gtk_widget_realize(GTK_WIDGET(menu));
+    /* get all the relative coordinates */
+    gtk_widget_get_allocation(widget, &a);
+    screen = gtk_widget_get_screen(widget);
+    gdk_window_get_device_position(gtk_widget_get_window(widget),
+                                   gdk_device_manager_get_client_pointer(
+                                        gdk_display_get_device_manager(
+                                            gdk_screen_get_display(screen))),
+                                   &x2, &y2, NULL);
+    gtk_widget_get_allocation(GTK_WIDGET(menu), &ma);
+    path = gtk_tree_path_new_from_indices(index, -1);
+    gtk_tree_view_get_cell_area(view, path, gtk_tree_view_get_column(view, 0), &cell);
+    gtk_tree_path_free(path);
+    /* position menu inside the cell if pointer isn't already in it */
+    if(x2 < cell.x || x2 > cell.x + cell.width)
+        x2 = cell.x + cell.width/2;
+    if(y2 < cell.y || y2 > cell.y + cell.height)
+        y2 = cell.y + cell.height - cell.height/8;
+    /* get absolute coordinate of parent window - we got coords relative to it */
+    gdk_window_get_origin(gtk_widget_get_parent_window(widget), x, y);
+    /* calculate desired position for menu */
+    *x += a.x + x2;
+    *y += a.y + y2;
+    /* limit coordinates so menu will be not positioned outside of screen */
+    mon = gdk_screen_get_monitor_at_point(screen, *x, *y);
+    /* get monitor geometry into the rectangle */
+    gdk_screen_get_monitor_geometry(screen, mon, &cell);
+    if(rtl) /* RTL */
+    {
+        x2 = cell.x + cell.width;
+        if (*x < cell.x + ma.width) /* out of monitor */
+            *x = MIN(*x + ma.width, x2); /* place menu right to cursor */
+        else
+            *x = MIN(*x, x2);
+    }
+    else /* LTR */
+    {
+        if (*x + ma.width > cell.x + cell.width) /* out of monitor */
+            *x = MAX(cell.x, *x - ma.width); /* place menu left to cursor */
+        else
+            *x = MAX(cell.x, *x); /* place menu right to cursor */
+    }
+    if (*y + ma.height > cell.y + cell.height) /* out of monitor */
+        *y = MAX(cell.y, *y - ma.height); /* place menu above cursor */
+    else
+        *y = MAX(cell.y, *y); /* place menu below cursor */
+}
+
+static void fm_places_item_popup(GtkWidget *widget, GtkTreeIter *it, guint32 time)
+{
+    if(!fm_places_model_iter_is_separator(model, it))
+    {
+        FmPlacesItem* item;
+        GtkMenu* menu;
+        GtkTreePath* path;
+        gint* indices;
+        gtk_tree_model_get(GTK_TREE_MODEL(model), it, FM_PLACES_MODEL_COL_INFO, &item, -1);
+        menu = GTK_MENU(place_item_get_menu(item, widget));
+        if(menu)
+        {
+            path = gtk_tree_model_get_path(GTK_TREE_MODEL(model), it);
+            indices = gtk_tree_path_get_indices(path);
+            gtk_menu_popup(menu, NULL, NULL, popup_position_func,
+                           GINT_TO_POINTER(indices[0]), 3, time);
+            gtk_tree_path_free(path);
+        }
+    }
+}
+
 static gboolean on_button_press(GtkWidget* widget, GdkEventButton* evt)
 {
     FmPlacesView* view = FM_PLACES_VIEW(widget);
     GtkTreePath* tp;
     GtkTreeViewColumn* col;
+    GtkTreeIter it;
     gboolean ret = GTK_WIDGET_CLASS(fm_places_view_parent_class)->button_press_event(widget, evt);
 
     gtk_tree_view_get_path_at_pos(&view->parent, evt->x, evt->y, &tp, &col, NULL, NULL);
@@ -805,27 +994,36 @@ static gboolean on_button_press(GtkWidget* widget, GdkEventButton* evt)
             activate_row(view, 2, tp);
             break;
         case 3: /* right click */
-            {
-                GtkTreeIter it;
-                if(gtk_tree_model_get_iter(GTK_TREE_MODEL(model), &it, tp)
-                    && !fm_places_model_iter_is_separator(model, &it) )
-                {
-                    FmPlacesItem* item;
-                    GtkMenu* menu;
-                    gtk_tree_model_get(GTK_TREE_MODEL(model), &it, FM_PLACES_MODEL_COL_INFO, &item, -1);
-                    menu = GTK_MENU(place_item_get_menu(item, widget));
-                    if(menu)
-                    {
-                        gtk_menu_popup(menu, NULL, NULL, NULL, NULL, 3, evt->time);
-                    }
-                }
-            }
+            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(model), &it, tp))
+                fm_places_item_popup(widget, &it, evt->time);
             break;
         }
     }
     return ret;
 }
 
+/* handle 'Menu' and 'Shift+F10' here */
+static gboolean on_key_press(GtkWidget *widget, GdkEventKey *evt)
+{
+    GtkTreeModel *model;
+    GtkTreeSelection *sel;
+    GtkTreeIter it;
+    int modifier = (evt->state & gtk_accelerator_get_default_mod_mask());
+
+    if((evt->keyval == GDK_KEY_Menu && !modifier) ||
+       (evt->keyval == GDK_KEY_F10 && modifier == GDK_SHIFT_MASK))
+    {
+        sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
+        if(gtk_tree_selection_get_selected(sel, &model, &it))
+        {
+            fm_places_item_popup(widget, &it, evt->time);
+            return TRUE;
+        }
+    }
+    /* let others do the job */
+    return GTK_WIDGET_CLASS(fm_places_view_parent_class)->key_press_event(widget, evt);
+}
+
 static void on_mount(GtkAction* act, gpointer user_data)
 {
     FmPlacesItem* item = (FmPlacesItem*)user_data;
@@ -862,17 +1060,55 @@ static void on_umount(GtkAction* act, gpointer user_data)
     if(mnt)
     {
         fm_unmount_mount(NULL, mnt, TRUE);
+        /* NOTE: the most probably FmPlacesView is destroyed at this point */
         g_object_unref(mnt);
     }
 }
 
+static GtkWindow *_get_gtk_window_from_action(GtkAction* act)
+{
+    /* FIXME: This is very dirty, but it's inevitable. :-( */
+    GSList* proxies = gtk_action_get_proxies(act);
+    GtkWidget *menu, *view;
+    menu = gtk_widget_get_parent(proxies->data);
+    if (menu == NULL || !GTK_IS_MENU(menu))
+        return NULL;
+    view = gtk_menu_get_attach_widget((GtkMenu*)menu);
+    return view ? GTK_WINDOW(gtk_widget_get_toplevel(view)) : NULL;
+}
+
 static void on_eject(GtkAction* act, gpointer user_data)
 {
     FmPlacesItem* item = (FmPlacesItem*)user_data;
     if(fm_places_item_get_type(item) == FM_PLACES_ITEM_VOLUME)
     {
-        /* FIXME: get the toplevel window here */
-        fm_eject_volume(NULL, fm_places_item_get_volume(item), TRUE);
+        fm_eject_volume(_get_gtk_window_from_action(act),
+                        fm_places_item_get_volume(item), TRUE);
+        /* NOTE: the most probably FmPlacesView is destroyed at this point */
+    }
+}
+
+static void on_format(GtkAction* act, gpointer user_data)
+{
+    FmPlacesItem* item = (FmPlacesItem*)user_data;
+    char *unix_path;
+
+    if (fm_config->format_cmd == NULL || fm_config->format_cmd[0] == 0)
+        return;
+    unix_path = g_volume_get_identifier(fm_places_item_get_volume(item),
+                                        G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
+    if (unix_path)
+    {
+        /* call fm_config->format_cmd for device */
+        g_debug("formatting %s ...", unix_path);
+        FmPath *path = fm_path_new_for_path(unix_path);
+        g_free(unix_path);
+        FmPathList *paths = fm_path_list_new();
+        fm_path_list_push_tail(paths, path);
+        fm_path_unref(path);
+        fm_launch_command_simple(_get_gtk_window_from_action(act), NULL, 0,
+                                 fm_config->format_cmd, paths);
+        fm_path_list_unref(paths);
     }
 }
 
@@ -887,8 +1123,8 @@ static void on_remove_bm(GtkAction* act, gpointer user_data)
 static void on_rename_bm(GtkAction* act, gpointer user_data)
 {
     FmPlacesItem* item = (FmPlacesItem*)user_data;
-    /* FIXME: we need to set a proper parent window for the dialog */
-    char* new_name = fm_get_user_input(NULL, _("Rename Bookmark Item"),
+    char* new_name = fm_get_user_input(_get_gtk_window_from_action(act),
+                                       _("Rename Bookmark Item"),
                                        _("Enter a new name:"),
                                        fm_places_item_get_bookmark_item(item)->name);
     if(new_name)
@@ -901,14 +1137,61 @@ static void on_rename_bm(GtkAction* act, gpointer user_data)
     }
 }
 
+static void on_move_bm_up(GtkAction* act, gpointer user_data)
+{
+    FmPlacesItem* item = (FmPlacesItem*)user_data;
+    int new_pos, sep_pos;
+    GtkTreeIter src_it, dest_it;
+    GtkTreePath *tp, *sep_tp;
+
+    if(!fm_places_model_get_iter_by_fm_path(model, &src_it,
+                                            fm_places_item_get_path(item)))
+        return; /* FIXME: print error message */
+    /* get index of the separator */
+    sep_tp = fm_places_model_get_separator_path(model);
+    sep_pos = gtk_tree_path_get_indices(sep_tp)[0];
+    tp = gtk_tree_model_get_path(GTK_TREE_MODEL(model), &src_it);
+    if(!gtk_tree_path_prev(tp) ||
+       (new_pos = gtk_tree_path_get_indices(tp)[0] - sep_pos - 1) < 0 ||
+       !gtk_tree_model_get_iter(GTK_TREE_MODEL(model), &dest_it, tp))
+        goto _end; /* cannot move it up */
+    gtk_list_store_move_before(GTK_LIST_STORE(model), &src_it, &dest_it);
+    /* reorder the bookmark item */
+    fm_bookmarks_reorder(fm_places_model_get_bookmarks(model), fm_places_item_get_bookmark_item(item), new_pos);
+_end:
+    gtk_tree_path_free(sep_tp);
+    gtk_tree_path_free(tp);
+}
+
+static void on_move_bm_down(GtkAction* act, gpointer user_data)
+{
+    FmPlacesItem* item = (FmPlacesItem*)user_data;
+    int new_pos, sep_pos;
+    GtkTreeIter src_it, dest_it;
+    GtkTreePath *tp, *sep_tp;
+
+    if(!fm_places_model_get_iter_by_fm_path(model, &src_it,
+                                            fm_places_item_get_path(item)))
+        return; /* FIXME: print error message */
+    /* get index of the separator */
+    sep_tp = fm_places_model_get_separator_path(model);
+    sep_pos = gtk_tree_path_get_indices(sep_tp)[0];
+    dest_it = src_it;
+    if(!gtk_tree_model_iter_next(GTK_TREE_MODEL(model), &dest_it))
+        goto _end; /* cannot move it down */
+    gtk_list_store_move_after(GTK_LIST_STORE(model), &src_it, &dest_it);
+    /* reorder the bookmark item */
+    tp = gtk_tree_model_get_path(GTK_TREE_MODEL(model), &src_it);
+    new_pos = gtk_tree_path_get_indices(tp)[0] - sep_pos - 1;
+    fm_bookmarks_reorder(fm_places_model_get_bookmarks(model), fm_places_item_get_bookmark_item(item), new_pos);
+    gtk_tree_path_free(tp);
+_end:
+    gtk_tree_path_free(sep_tp);
+}
+
 static void on_empty_trash(GtkAction* act, gpointer user_data)
 {
-    /* FIXME: This is very dirty, but it's inevitable. :-( */
-    GSList* proxies = gtk_action_get_proxies(act);
-    GtkWidget* menu_item = proxies->data ? GTK_WIDGET(proxies->data) : NULL;
-    GtkWidget* menu = gtk_widget_get_parent(menu_item);
-    GtkWidget* view = gtk_menu_get_attach_widget(GTK_MENU(menu));
-    fm_empty_trash(view ? GTK_WINDOW(gtk_widget_get_toplevel(view)) : NULL);
+    fm_empty_trash(_get_gtk_window_from_action(act));
 }
 
 #if !GTK_CHECK_VERSION(3, 0, 0)
@@ -920,6 +1203,49 @@ static void on_set_scroll_adjustments(GtkTreeView* view, GtkAdjustment* hadj, Gt
 }
 #endif
 
+static void fm_places_view_set_property(GObject *object,
+                                        guint prop_id,
+                                        const GValue *value,
+                                        GParamSpec *pspec)
+{
+    FmPlacesView *view = FM_PLACES_VIEW(object);
+    const char *home_dir;
+
+    switch( prop_id )
+    {
+    case PROP_HOME_DIR:
+        home_dir = g_value_get_string(value);
+        if (home_dir && (!*home_dir || strcmp(home_dir, fm_get_home_dir()) == 0))
+            home_dir = NULL;
+        g_free(view->home_dir);
+        view->home_dir = g_strdup(home_dir);
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+        break;
+    }
+}
+
+static void fm_places_view_get_property(GObject *object,
+                                        guint prop_id,
+                                        GValue *value,
+                                        GParamSpec *pspec)
+{
+    FmPlacesView *view = FM_PLACES_VIEW(object);
+
+    switch( prop_id ) {
+    case PROP_HOME_DIR:
+        if (view->home_dir)
+            g_value_set_string(value, view->home_dir);
+        else
+            g_value_set_string(value, fm_get_home_dir());
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+        break;
+    }
+}
+
 static void fm_places_view_class_init(FmPlacesViewClass *klass)
 {
     GObjectClass *g_object_class;
@@ -930,6 +1256,7 @@ static void fm_places_view_class_init(FmPlacesViewClass *klass)
     g_object_class->finalize = fm_places_view_finalize;
 
     widget_class = GTK_WIDGET_CLASS(klass);
+    widget_class->key_press_event = on_key_press;
     widget_class->button_press_event = on_button_press;
     widget_class->button_release_event = on_button_release;
     widget_class->drag_motion = on_drag_motion;
@@ -943,6 +1270,26 @@ static void fm_places_view_class_init(FmPlacesViewClass *klass)
     tv_class->set_scroll_adjustments = on_set_scroll_adjustments;
 #endif
 
+    g_object_class->get_property = fm_places_view_get_property;
+    g_object_class->set_property = fm_places_view_set_property;
+
+    /**
+     * FmPlacesView:home-dir-path:
+     *
+     * The #FmPlacesView:home-dir-path property defines which path will
+     * be used on Home item activation. Value of %NULL resets it to the
+     * default.
+     *
+     * Since: 1.2.0
+     */
+    g_object_class_install_property(g_object_class,
+                                    PROP_HOME_DIR,
+                                    g_param_spec_string("home-dir-path",
+                                                        "Home item directory",
+                                                        "What directory path will be used for Home item",
+                                                        NULL,
+                                                        G_PARAM_READWRITE));
+
     /**
      * FmPlacesView::chdir:
      * @view: a view instance that emitted the signal
@@ -963,5 +1310,27 @@ static void fm_places_view_class_init(FmPlacesViewClass *klass)
                      g_cclosure_marshal_VOID__UINT_POINTER,
                      G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_POINTER);
 
+    /**
+     * FmPlacesView::item-popup:
+     * @view: a view instance that emitted the signal
+     * @ui: the #GtkUIManager using to create the menu
+     * @act_grp: (#GtkActionGroup *) the menu actions group
+     * @fi: (#FmFileInfo *) the item where menu popup is activated
+     *
+     * The #FmPlacesView::item-popup signal is emitted when context menu
+     * is created for any directory in the view. Handler can modify the
+     * menu by adding or removing elements.
+     *
+     * Since: 1.2.0
+     */
+    signals[ITEM_POPUP] =
+        g_signal_new("item-popup",
+                     G_TYPE_FROM_CLASS(klass),
+                     G_SIGNAL_RUN_LAST,
+                     G_STRUCT_OFFSET(FmPlacesViewClass, item_popup),
+                     NULL, NULL,
+                     fm_marshal_VOID__OBJECT_OBJECT_POINTER,
+                     G_TYPE_NONE, 3, G_TYPE_OBJECT, G_TYPE_OBJECT, G_TYPE_POINTER);
+
     tree_model_row_atom = gdk_atom_intern_static_string("GTK_TREE_MODEL_ROW");
 }
