@@ -1,6 +1,7 @@
 //      fm-side-pane.c
 //
 //      Copyright 2011 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
+//      Copyright 2013-2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
 //
 //      This program is free software; you can redistribute it and/or modify
 //      it under the terms of the GNU General Public License as published by
@@ -22,13 +23,17 @@
  * @short_description: A widget for side pane displaying
  * @title: FmSidePane
  *
- * @include: libfm/fm-side-pane.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmSidePane widget displays side pane for fast navigation across
  * places.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
+#define FM_DISABLE_SEAL
 
 #include "fm-side-pane.h"
 
@@ -294,9 +299,6 @@ void fm_side_pane_chdir(FmSidePane* sp, FmPath* path)
 
 static void on_places_chdir(FmPlacesView* view, guint button, FmPath* path, FmSidePane* sp)
 {
-//    g_signal_handlers_block_by_func(win->dirtree_view, on_dirtree_chdir, win);
-//    fm_main_win_chdir(win, path);
-//    g_signal_handlers_unblock_by_func(win->dirtree_view, on_dirtree_chdir, win);
     if(sp->cwd)
         fm_path_unref(sp->cwd);
     sp->cwd = fm_path_ref(path);
@@ -305,15 +307,19 @@ static void on_places_chdir(FmPlacesView* view, guint button, FmPath* path, FmSi
 
 static void on_dirtree_chdir(FmDirTreeView* view, guint button, FmPath* path, FmSidePane* sp)
 {
-//    g_signal_handlers_block_by_func(win->places_view, on_places_chdir, win);
-//    fm_main_win_chdir(win, path);
-//    g_signal_handlers_unblock_by_func(win->places_view, on_places_chdir, win);
     if(sp->cwd)
         fm_path_unref(sp->cwd);
     sp->cwd = fm_path_ref(path);
     g_signal_emit(sp, signals[CHDIR], 0, button, path);
 }
 
+static void on_item_popup(GtkWidget* view, GtkUIManager* ui,
+                          GtkActionGroup* act_grp, FmFileInfo* file,
+                          FmSidePane* sp)
+{
+    sp->update_popup(sp, ui, act_grp, file, sp->popup_user_data);
+}
+
 static void fm_side_pane_dispose(GObject *object)
 {
     FmSidePane *sp;
@@ -344,9 +350,13 @@ static void fm_side_pane_dispose(GObject *object)
         switch(sp->mode)
         {
         case FM_SP_PLACES:
+            if (sp->update_popup)
+                g_signal_handlers_disconnect_by_func(sp->view, on_item_popup, sp);
             g_signal_handlers_disconnect_by_func(sp->view, on_places_chdir, sp);
             break;
         case FM_SP_DIR_TREE:
+            if (sp->update_popup)
+                g_signal_handlers_disconnect_by_func(sp->view, on_item_popup, sp);
             g_signal_handlers_disconnect_by_func(sp->view, on_dirtree_chdir, sp);
             break;
         default: ; /* other values are impossible, otherwise it's a bug */
@@ -372,7 +382,9 @@ static void init_dir_tree(FmSidePane* sp)
         fm_file_info_job_add(job, fm_path_get_root());
 
         /* FIXME: maybe it's cleaner to use run_async here? */
+        GDK_THREADS_LEAVE();
         fm_job_run_sync_with_mainloop(FM_JOB(job));
+        GDK_THREADS_ENTER();
 
         dir_tree_model = fm_dir_tree_model_new();
         for(l = fm_file_info_list_peek_head_link(job->file_infos); l; l = l->next)
@@ -401,10 +413,14 @@ void fm_side_pane_set_mode(FmSidePane* sp, FmSidePaneMode mode)
 {
     if(mode == sp->mode)
         return;
-    sp->mode = mode;
 
     if(sp->view)
+    {
+        if (sp->update_popup)
+            g_signal_handlers_disconnect_by_func(sp->view, on_item_popup, sp);
         gtk_widget_destroy(sp->view);
+    }
+    sp->mode = mode;
 
     switch(mode)
     {
@@ -435,6 +451,8 @@ void fm_side_pane_set_mode(FmSidePane* sp, FmSidePaneMode mode)
         /* not implemented */
         return;
     }
+    if (sp->update_popup)
+        g_signal_connect(sp->view, "item-popup", G_CALLBACK(on_item_popup), sp);
     gtk_widget_show(sp->view);
     gtk_container_add(GTK_CONTAINER(sp->scroll), sp->view);
 
@@ -474,3 +492,213 @@ GtkWidget* fm_side_pane_get_title_bar(FmSidePane* sp)
 {
     return sp->title_bar;
 }
+
+/**
+ * fm_side_pane_set_popup_updater
+ * @sp: a widget to set
+ * @update_popup: (allow-none): a callback to update popup
+ * @user_data: user data supplied for callback
+ *
+ * Sets up the callback to update context menu on any item in the sidebar
+ * widget.
+ *
+ * Since: 1.2.0
+ */
+void fm_side_pane_set_popup_updater(FmSidePane* sp,
+                                    FmSidePaneUpdatePopup update_popup,
+                                    gpointer user_data)
+{
+    gboolean was_set = (sp->update_popup != NULL);
+
+    sp->update_popup = update_popup;
+    sp->popup_user_data = user_data;
+    if (sp->view == NULL)
+        return; /*  nothing to do yet */
+    if (was_set)
+    {
+        switch (sp->mode)
+        {
+        case FM_SP_PLACES:
+        case FM_SP_DIR_TREE:
+            if (update_popup == NULL)
+                g_signal_handlers_disconnect_by_func(sp->view, on_item_popup, sp);
+            break;
+        default: ;
+        }
+    }
+    else if (update_popup)
+    {
+        switch (sp->mode)
+        {
+        case FM_SP_PLACES:
+        case FM_SP_DIR_TREE:
+            g_signal_connect(sp->view, "item-popup", G_CALLBACK(on_item_popup), sp);
+            break;
+        default: ;
+        }
+    }
+}
+
+/**
+ * fm_side_pane_get_mode_name
+ * @mode: mode to convert
+ *
+ * Retrieves the name of the specified side pane @mode, or %NULL if @mode
+ * is invalid. The name of mode may be used for config save or another
+ * similar purpose. Returned data are owned by implementation and should
+ * be not freed by caller.
+ *
+ * Returns: the name associated with @mode.
+ *
+ * Since: 1.2.0
+ */
+const char *fm_side_pane_get_mode_name(FmSidePaneMode mode)
+{
+    switch (mode)
+    {
+        case FM_SP_PLACES:
+            return "places";
+        case FM_SP_DIR_TREE:
+            return "dirtree";
+        default:
+            return NULL;
+    }
+}
+
+/**
+ * fm_side_pane_get_mode_by_name
+ * @str: the name of mode
+ *
+ * Finds mode which have an associated name equal to @str.
+ *
+ * Returns: mode or FM_SP_NONE if @str specifies invalid mode.
+ *
+ * Since: 1.2.0
+ */
+FmSidePaneMode fm_side_pane_get_mode_by_name(const char *str)
+{
+    if (str == NULL)
+        return FM_SP_NONE;
+    if (strcmp(str, "places") == 0)
+        return FM_SP_PLACES;
+    if (strcmp(str, "dirtree") == 0)
+        return FM_SP_DIR_TREE;
+    return FM_SP_NONE;
+}
+
+/**
+ * fm_side_pane_get_n_modes
+ *
+ * Tests how many view modes are known to create #FmSidePane widget. The
+ * returned value is the highest FmSidePaneMode that can be used due to
+ * historical reasons since FmSidePaneMode equal to 0 is invalid.
+ *
+ * Returns: number of known modes for side pane.
+ *
+ * Since: 1.2.0
+ */
+gint fm_side_pane_get_n_modes(void)
+{
+    return (FM_SP_DIR_TREE + 1);
+}
+
+/**
+ * fm_side_pane_get_mode_label
+ * @mode: the view mode
+ *
+ * Retrieves label for @mode which can be used in menus. Returned
+ * data should not be freed by caller.
+ *
+ * Returns: desription or %NULL if @mode is invalid.
+ *
+ * Since: 1.2.0
+ */
+const char *fm_side_pane_get_mode_label(FmSidePaneMode mode)
+{
+    switch (mode)
+    {
+        case FM_SP_PLACES:
+            return _("Places");
+        case FM_SP_DIR_TREE:
+            return _("Directory Tree");
+        default:
+            return NULL;
+    }
+}
+
+/**
+ * fm_side_pane_get_mode_tooltip
+ * @mode: the view mode
+ *
+ * Retrieves tooltip for @mode which can be used in menus. Returned
+ * data should not be freed by caller.
+ *
+ * Returns: desription or %NULL if @mode is invalid.
+ *
+ * Since: 1.2.0
+ */
+const char *fm_side_pane_get_mode_tooltip(FmSidePaneMode mode)
+{
+    switch (mode)
+    {
+        case FM_SP_PLACES:
+            return _("Shows list of common places, devices, and bookmarks in sidebar");
+        case FM_SP_DIR_TREE:
+            return _("Shows tree of directories in sidebar");
+        default:
+            return NULL;
+    }
+}
+
+/**
+ * fm_side_pane_set_show_hidden
+ * @sp: a widget to apply
+ * @show_hidden: %TRUE to show hidden files in side pane
+ *
+ * Changes visibility of hidden files in side pane @sp.
+ *
+ * Returns: %TRUE if current side pane mode accepts this change.
+ *
+ * Since: 1.2.0
+ */
+gboolean fm_side_pane_set_show_hidden(FmSidePane *sp, gboolean show_hidden)
+{
+    GObjectClass *klass;
+    GParamSpec *spec;
+
+    if (sp->view == NULL)
+        return FALSE;
+    klass = G_OBJECT_GET_CLASS(sp->view);
+    spec = g_object_class_find_property(klass, "show-hidden");
+    if (spec == NULL || spec->value_type != G_TYPE_BOOLEAN)
+        return FALSE; /* isn't supported by view */
+    g_object_set(sp->view, "show-hidden", show_hidden, NULL);
+    return TRUE;
+}
+
+/**
+ * fm_side_pane_set_home_dir
+ * @sp: a widget to apply
+ * @home_dir: a new path to set
+ *
+ * Changes path used by side pane view for "Home" items if view mode has
+ * any such items.
+ *
+ * Returns: %TRUE if current side pane mode accepts this change.
+ *
+ * Since: 1.2.0
+ */
+gboolean fm_side_pane_set_home_dir(FmSidePane *sp, const char *home_dir)
+{
+    GObjectClass *klass;
+    GParamSpec *spec;
+
+    if (sp->view == NULL)
+        return FALSE;
+    klass = G_OBJECT_GET_CLASS(sp->view);
+    spec = g_object_class_find_property(klass, "home-dir-path");
+    if (spec == NULL || spec->value_type != G_TYPE_STRING)
+        return FALSE; /* isn't supported by view */
+    g_object_set(sp->view, "home-dir-path", home_dir, NULL);
+    return TRUE;
+}
